'''
standardize this package's version string representation

This is an alternative to using the versioneer package
which had some problems, such as:
https://github.com/prjemian/spec2nexus/issues/61#issuecomment-246021134

USAGE:

    In the package's ``__init__.py`` file, place these commands::

        __package_name__ = u'package_name'
        
        from _version import git_release
        __version__ = u'1.2.3'
        __release__ = git_release(__package_name__, __version__)

'''

DEVELOPER_TEST_STRING = '__developer_testing__'

def git_release(package, version='release_undefined'):
    '''
    get the version string from the current git tag and commit info, if available
    '''
    release = version
    try:
        import os, subprocess
        
        # First, verify that we are in the development directory of this package.
        # Package name must match the name of the directory containing this file.
        # Otherwise, it is possible that the current working directory might have
        # a valid response to the "git describe" command and return the wrong version string.
        path = os.path.dirname(__file__)
        dirname = os.path.split(path)[-1]
        if package not in (dirname, DEVELOPER_TEST_STRING):
            raise ValueError
        
        git_command = 'git describe'.split()
        p = subprocess.Popen(git_command,
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, _err = p.communicate()
        if out:
            release = out.decode().strip()
    except Exception as _exc:
        pass
    return release


if __name__ == '__main__':
    print(git_release(DEVELOPER_TEST_STRING, 'git_release_string'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

'''
validate NeXus NXDL and HDF5 data files

.. autosummary::
   
   ~validate_xml
   ~NxdlPattern
   ~CustomNxdlPattern
   ~Data_File_Validator

.. rubric:: CHECKLIST

These are considerata for the validation of NeXus HDF5 data files.
Compare these validation steps with rules and documentation
in the NeXus manual and the XML Schema files (``nxdl.xsd`` and ``nxdlTypes.xsd``).
Checkboxes indicate which steps have been implemented in code below.

* [x] make a list of all address nodes in the file to be evaluated
* [x] attributes are also in this list
* [x] use a structure to hold results for each node

.. rubric:: File

#. [x] verify attributes
#. [x] verify file level as group using NX_class = NXroot
#. [ ] identify any objects at root level that are not in NXroot (which is OK)
#. [x] verify default plot identified

    #. [x] version 1
    #. [x] version 2
    #. [x] version 3
    #. [x] version 3+niac2014

.. rubric:: Groups

#. [x] compare name with pattern *validItemName*
#. [x] determine NX_class, if any
#. [x] verify NX_class in nxdl_dict
#. [ ] is name flexible?
#. [ ] What to do with NXDL symbol tables?
#. [x] observe attribute: minOccurs
#. [ ] observe attribute: maxOccurs
#. [ ] check for items defined by NX_class
#. [ ] check for items required by NX_class
#. [ ] check for items not defined by NX_class
#. [x] observe NXDL specification: ignoreExtraGroups
#. [x] observe NXDL specification: ignoreExtraFields
#. [x] observe NXDL specification: ignoreExtraAttributes
#. [x] validate any attributes
#. [x] validate any links
#. [x] validate any fields

.. rubric:: Links

#. [x] compare name with pattern *validItemName*
#. [ ] is name flexible?
#. [x] is target attribute defined?
#. [x] is target address absolute?
#. [x] does target address exist?
#. [ ] construct NX classpath from target
#. [ ] compare NX classpath with NXDL specification

.. rubric:: Fields

#. [x] compare name with pattern
#. [x] is name flexible?
#. [x] observe attribute: minOccurs
#. [x] is units attribute defined?
#. [x] check units are consistent against NXDL
#. [x] check data shape against NXDL
#. [x] check data type against NXDL
#. [x] check for attributes defined by NXDL
#. [x] check AXISNAME_indices are each within signal data rank

.. rubric:: Attributes

#. [x] compare name with pattern
#. [ ] check data type against NXDL
#. [ ] check nxdl.xsd for how to handle these attributes regarding finding.WARN
'''

import collections
import h5py
import lxml.etree
import numpy
import os
import re

import __init__
import cache
import finding
import h5structure
import nxdlstructure

# TODO: issue #14: http://download.nexusformat.org/doc/html/search.html?q=warning&check_keywords=yes&area=default

__url__ = 'http://punx.readthedocs.org/en/latest/validate.html'


def validate_xml(xml_file_name):
    '''
    validate an NXDL XML file against an XML Schema file

    :param str xml_file_name: name of XML file
    '''
    xml_tree = lxml.etree.parse(xml_file_name)
    xsd = cache.get_XML_Schema()
    try:
        result = xsd.assertValid(xml_tree)
    except lxml.etree.DocumentInvalid as exc:
        msg = 'DocumentInvalid:\n'
        msg += 'file: ' + xml_file_name + '\n'
        msg += str(exc)
        raise Exception(msg)
    return result


class NxdlPattern(object):
    '''
    common regular expression pattern for validation
    
    :param obj parent: instance of :class:`Data_File_Validator`
    :param str pname: pattern identifying name
    :param str xpath_str: XPath search string, expect list of length = 1
    '''
    
    def __init__(self, parent, pname, xpath_str):
        self.name = pname
        self.xpath_str = xpath_str
        rules = cache.get_nxdl_xsd()

        r = rules.xpath(xpath_str, namespaces=parent.ns)

        if r is None or len(r) != 1:
            msg = 'could not read *' + pname + '* from *nxdl.xsd*'
            raise ValueError(msg)

        self.regexp_pattern_str = r[0].attrib.get('value', None)
        self.re_obj = re.compile('^' + self.regexp_pattern_str + '$')
    
    def match(self, text):
        '''regular expression search'''
        return self.re_obj.match(text)


class CustomNxdlPattern(NxdlPattern):
    '''
    custom regular expression pattern for validation
    
    :param obj parent: instance of :class:`Data_File_Validator`
    :param str pname: pattern identifying name
    :param str regexp_pattern_str: regular expression to match
    '''
    
    def __init__(self, parent, pname, regexp_pattern_str):
        self.name = pname
        self.xpath_str = None

        self.regexp_pattern_str = regexp_pattern_str
        self.re_obj = re.compile('^' + self.regexp_pattern_str + '$')
    
    def match(self, text):
        '''regular expression search'''
        return self.re_obj.match(text)


class Data_File_Validator(object):
    '''
    manage the validation of a NeXus HDF5 data file
    '''
    
    def __init__(self, fname):
        if not os.path.exists(fname):
            raise __init__.FileNotFound(fname)
        self.fname = fname

        self.findings = []      # list of Finding() instances
        self.addresses = collections.OrderedDict()     # dictionary of all HDF5 address nodes in the data file

        self.ns = cache.NX_DICT
        self.nxdl_rules = nxdlstructure.get_nxdl_rules()
        self.get_data_types()
        self.nxdl_dict = nxdlstructure.get_NXDL_specifications()

        try:
            self.h5 = h5py.File(fname, 'r')
        except IOError:
            raise __init__.HDF5_Open_Error(fname)
        self._init_patterns()
    
    def _init_patterns(self):
        self.patterns = {}
        for item in ('validItemName', 'validNXClassName', 
                     'validTargetName'):
            xps = '//*[@name="' # XPath String query
            xps += item
            xps += '"]/xs:restriction/xs:pattern'
            self.patterns[item] = NxdlPattern(self, item, xps)

        # strict match: [a-z_][a-z\d_]*
        # flexible match: [A-Za-z_][\w_]*  but gets finding.WARN per manual
        # advisory changed to finding.NOTE
        p = CustomNxdlPattern(self, 'validItemName-strict', r'[a-z_][a-z0-9_]*')
        self.patterns[p.name] = p
        self.__unique_findings__ = {}
    
    def get_data_types(self):
        '''
        generate dictionary of acceptable Python data types, based on NeXus data type keys
        '''
        # Is there a better way to define these?  Using nxdlTypes.xsd?
        # TODO: #21 : augment from self.nxdl_rules.nxdlTypes
        
        self.data_types = {
            'NX_CHAR': (str, unicode, numpy.string_, numpy.ndarray),
            'NX_UINT': (numpy.uint, numpy.uint8, numpy.uint16, numpy.uint32, numpy.uint64),
            'NX_INT':  (int, numpy.int, numpy.int8, numpy.int16, numpy.int32, numpy.int64),
            'NX_FLOAT':  (float, numpy.float, numpy.float16, numpy.float32, numpy.float64),
            'NX_BINARY': (None, ),     # #FIXME:
            'NX_BOOLEAN': (None, ),     # FIXME:
        }
        # definitions dependent on other definitions
        # (can add the lists together as needed)
        self.data_types['NX_INT']    += self.data_types['NX_UINT']
        self.data_types['NX_POSINT'] = self.data_types['NX_INT']    # need to restrict this
        self.data_types['NX_NUMBER'] = self.data_types['NX_INT'] + self.data_types['NX_FLOAT']
        self.data_types['ISO8601']   = self.data_types['NX_CHAR']
        self.data_types['NX_DATE_TIME']   = self.data_types['NX_CHAR']
        
    def validate(self):
        '''
        start the validation process from the file root
        '''
        self.validate_HDF5_group(self.h5)
        self.validate_default_plot()
        
    def validate_HDF5_group(self, group):
        '''
        review the HDF5 group: group

        :param obj group: instance of h5py.File of h5py.Group
        
        Verify that items presented in data file are valid.
        '''
        nx_class_name = self.get_hdf5_attribute(group, 'NX_class', report=True)
        if nx_class_name is None:
            if isinstance(group, h5py.File):
                nx_class_name = 'NXroot'
                msg = 'file root (assumed): NXroot'
                self.new_finding('@NX_class', group.name, finding.OK, msg)
            else:
                # TODO: validNXClassName - use this instead
                self.validate_item_name(group.name)
                msg = 'no @NX_class attribute, not a NeXus group'
                self.new_finding('@NX_class', group.name, finding.NOTE, msg)
                return  # evaluate any further?
        else:
            self.validate_item_name(group.name)

            aname = group.name + '@NX_class'
            t = nx_class_name in self.nxdl_dict
            f = finding.TF_RESULT[t]
            msg = {True: 'known: ', False: 'unknown: '}[t] + nx_class_name
            self.new_finding('@NX_class', aname, f, msg)
        
        nx_class_object = self.nxdl_dict.get(nx_class_name)
        if nx_class_object is None:
            msg = 'ignoring content in group not defined by NeXus'
            self.new_finding('not NeXus group', group.name, finding.UNUSED, msg)
            return  # ignore content in HDF5 groups that are not NeXus classes

        nx_class_defaults = nx_class_object.attributes['defaults']
        
        for k in group.attrs.keys():   # review the group's attributes
            if k in ('NX_class', ):
                pass    # handled elsewhere
            else:
                aname = group.name + '@' + k
                if not nx_class_defaults['ignoreExtraAttributes']:
                    # TODO: don't validate attribute names defined in NXDL rules!
                    self.validate_item_name(aname)
                if k == 'default' and nx_class_name in ('NXroot', 'NXentry', 'NXsubentry'):
                    target = self.get_hdf5_attribute(group, 'default')
                    t = target in group
                    f = finding.TF_RESULT[t]
                    msg = {True: 'exists: ', False: 'does not exist: '}[t] + target
                    self.new_finding('default plot group', aname, f, msg)

        for child_name in group:           # review the group's children
            try:
                child = group[child_name]
            except KeyError as _exc:
                filename = self.missing_file_link(_exc.message)
                if filename is not None:
                    title = 'external file link'
                    msg = 'missing file: ' + filename
                    self.new_finding(title, group.name+'/'+child_name, finding.ERROR, msg)
                continue
            if h5structure.isNeXusLink(child):
                if nx_class_defaults['ignoreExtraGroups']:
                    title = nx_class_name+'@ignoreExtraGroups'
                    msg = 'link ignored per NXDL specification'
                    self.new_finding(title, child.name, finding.UNUSED, msg)
                else:
                    self.validate_NeXus_link(child, group)
            elif h5structure.isHdf5Group(child):
                if nx_class_defaults['ignoreExtraGroups']:
                    title = nx_class_name+'@ignoreExtraGroups'
                    msg = 'subgroup ignored per NXDL specification'
                    self.new_finding(title, child.name, finding.UNUSED, msg)
                else:
                    self.validate_HDF5_group(child)
            elif h5structure.isHdf5Dataset(child):
                if nx_class_defaults['ignoreExtraFields']:
                    title = nx_class_name+'@ignoreExtraFields'
                    msg = 'field ignored per NXDL specification'
                    self.new_finding(title, child.name, finding.UNUSED, msg)
                else:
                    self.validate_HDF5_dataset(child, group)
            else:
                msg = 'unexpected: ' + child.name
                raise ValueError(msg)
        
        self.validate_NXDL_specification(group, nx_class_name)
        # FIXME: need special handling for application definitions
#         if nx_class_name in ('NXsubentry', 'NXentry') and 'definition' in group:
#             # application definition masquerading as NXentry or NXsubentry
#             app_def_name = group['definition'][0]
#             self.validate_NXDL_specification(group, app_def_name)
   
    def validate_HDF5_dataset(self, dataset, group):
        '''
        review the HDF5 dataset: dataset

        :param obj dataset: instance of h5py.Dataset
        :param obj group: instance of h5py.Group or h5py.File, needed to check against NXDL
        '''
        self.validate_item_name(dataset.name)
        field_rules = self.nxdl_rules.nxdl.children['field']
        nx_class_name = self.get_hdf5_attribute(group, 'NX_class')
        nx_class_object = self.nxdl_dict.get(nx_class_name)

        for k in dataset.attrs.keys():   # review the dataset's attributes
            aname = dataset.name + '@' + k
            self.validate_item_name(aname)
            v = self.get_hdf5_attribute(dataset, k, report=True)
            if k in field_rules.attrs:
                rules = field_rules.attrs[k]
                if len(rules.enum) > 0:
                    t = v in rules.enum
                    f = {True: finding.OK, False: finding.WARN}[t]
                    msg = 'value=' + v
                    if t:
                        msg += ' :recognized'
                    else:
                        msg += ' : expected one of these: ' + '|'.join(rules.enum)
                    self.new_finding('enumeration: @' + k, aname, f, msg)
            else:
                if k not in ('target',):  # link target attribute checked elsewhere
                    if nx_class_object is not None: # only check if NXDL exists
                        if nx_class_object.attributes['defaults']['ignoreExtraAttributes']:
                            msg = 'attribute ignored per NXDL specification'
                            self.new_finding(nx_class_name + '@ignoreExtraAttributes', aname, finding.NOTE, msg)
                        else:
                            msg = 'attribute not defined in NXDL'
                            self.new_finding(nx_class_name + '@' + k, aname, finding.NOTE, msg)
        
        self.validate_numerical_dataset(dataset, group)

        # check the type of this field
        # https://github.com/prjemian/punx/blob/b595fdf9910dbab113cfe8febbb37e6c5b48d74f/src/punx/validate.py#L761

        # review the dataset's content
        nx_class_name = self.get_hdf5_attribute(group, 'NX_class')
        if nx_class_name in self.nxdl_dict:
            nx_class = self.nxdl_dict[nx_class_name]
            rules = nx_class.fields.get(dataset.name.split('/')[-1])
            if rules is not None:
                if len(rules.enum) > 0:
                    pass    # TODO:
                defaults = rules.attributes['defaults']
                nx_type = defaults['type']      # TODO: check for this
                minO = defaults['minOccurs']    # TODO: check for this
                # in either case, validation of maxOccurs for datasets is not informative
                # HDF5 will not allow more than one instance of a name within a group
                # maxOccurs: is either 1 or, if name is flexible, unbounded

                isSpecifiedName = defaults['nameType'] == 'specified'
                f = finding.OK
                msg = 'name is ' + {True: 'specified', False: 'flexible'}[isSpecifiedName]
                self.new_finding('@nameType', dataset.name, f, msg)
                        
    def validate_NeXus_link(self, link, group):
        '''
        review the NeXus link: link
        
        :param obj link: instance of h5py.Group or h5py.Dataset
        :param obj group: instance of h5py.Group, needed to check against NXDL
        '''
        #: TODO: validTargetName  - use this instead
        self.validate_item_name(link.name)

        target = self.get_hdf5_attribute(link, 'target', report=True)
        if target is not None:
            aname = link.name + '@target'
            target_exists = target in self.h5
            f = finding.TF_RESULT[target_exists]
            msg = {True: target, False: 'does not exist'}[target_exists]
            self.new_finding('link target exists', aname, f, msg)
        else:
            self.new_finding('link', link.name, finding.ERROR, 'no target')
        
    def validate_NXDL_specification(self, group, nx_class_name):
        '''
        validate the group with the NXDL specification

        :param obj group: instance of h5py.Group or h5py.File
        :param str nx_class_name: name of a NeXus NXDL class
        
        Verify that items specified in NXDL file are present in the data file.
        '''
        nx_class_object = self.nxdl_dict.get(nx_class_name)
        if nx_class_object is None:
            return

        msg = 'validate with ' + nx_class_name + ' specification (incomplete)'
        self.new_finding('NXDL review', group.name, finding.TODO, msg)
        
        # specified group attributes are handled elsewhere
        # group_defaults = nx_class_object.attributes['defaults']

        # validate provided, required, and optional fields
        for field_name, rules in nx_class_object.fields.items():
            self.validate_NXDL_field_specification(field_name, group, rules)

        # validate provided, required, and optional groups (recursive as directed)
        for subgroup_name, rules in nx_class_object.groups.items():
            self.validate_NXDL_group_specification(subgroup_name, group, rules)
        
    def validate_NXDL_group_specification(self, subgroup_name, group, rules):
        '''
        validate the group/subgroup with the NXDL specification

        :param str subgroup_name: name of subgroup in group
        :param obj group: instance of h5py.Group or h5py.File
        :param obj rules: instance of nxdlstructure.NX_group
        
        Verify this HDF5 group conforms to the NXDL specification
        '''
        nx_class_name = self.get_hdf5_attribute(group, 'NX_class')
        defaults = rules.attributes['defaults']
        target_exists = subgroup_name in group

        deprecated = defaults['deprecated']
        if deprecated is not None:
            if target_exists:
                obj = group[subgroup_name]
                nm = '/'.join(rules.NX_class, subgroup_name) + '@deprecated'
                self.new_finding(nm, obj.name, finding.NOTE, deprecated)

        minO = defaults['minOccurs']
        maxO = defaults['maxOccurs']
        if int(minO) > 0:
            if defaults['name'] is None:
                matches = [node for node in group.values() if h5structure.isNeXusGroup(node, rules.NX_class)]
                if len(matches) < int(minO):
                    nm = group.name
                    m = 'must have at least ' + str(minO) + ' group: ' + rules.NX_class 
                    f = finding.WARN
                    self.new_finding(rules.NX_class+' required group', nm, f, m)
            else:
                nm = group.name + '/' + subgroup_name
                f = {True: finding.OK, False: finding.WARN}[target_exists]
                m = rules.NX_class + {True: ' found', False: ' not found'}[target_exists]
                self.new_finding(rules.NX_class+' required group', nm, f, m)
        if maxO == 'unbounded':
            pass
        # TODO: what else?
        
    def validate_NXDL_field_specification(self, field_name, group, rules):
        '''
        validate the group/field with the NXDL specification

        :param str field_name: name of field in group
        :param obj group: instance of h5py.Group or h5py.File
        :param obj rules: instance of nxdlstructure.NX_field
        
        Verify this HDF5 field conforms to the NXDL specification
        '''
        nx_class_name = self.get_hdf5_attribute(group, 'NX_class')
        defaults = rules.attributes['defaults']
        nx_type = self.data_types[defaults['type']]
        target_exists = field_name in group
        if target_exists:
            try:
                dataset = group[field_name]
            except KeyError as _exc:
                filename = self.missing_file_link(_exc.message)
                if filename is not None:
                    title = 'external file link'
                    msg = 'missing file: ' + filename
                    self.new_finding(title, group.name+'/'+field_name, finding.ERROR, msg)
                    return
        else:
            dataset = None
        nm = '/'.join((nx_class_name, field_name))

        # check the attributes specified in NXDL rules
        for k, attr in rules.attributes['nxdl.xsd'].items():
            if k in ('minOccurs maxOccurs name nameType type units'.split()):
                pass
            elif target_exists:
                if k == 'deprecated':
                    m = defaults['deprecated']
                    if m is not None:
                        self.new_finding(nm+'@deprecated', dataset.name, finding.NOTE, m)
                elif k in dataset.attrs:
                    aname = dataset.name + '@' + k
                    ttl = 'NXDL attribute type: '
                    ttl += '/'.join((nx_class_name, field_name))
                    ttl += '@' + k
                    v = dataset.attrs[k]
                     
                    # check type against NXDL
                    attr_type = attr.type.split(':')[-1]    # strip off XML namespace prefix, if found
                    t = type(v) in self.data_types[attr_type]
                    f = {True: finding.OK, False: finding.WARN}[t]
                    m = str(type(v)) + ' : ' + attr_type
                    self.new_finding(ttl, aname, f, m)
                     
                    # check if value matches enumeration
                    if len(attr.enum) > 0:
                        t = v in attr.enum
                        f = {True: finding.OK, False: finding.WARN}[t]
                        m = str(v)
                        if t:
                            m += ': expected'
                        else:
                            m += ': not in list: ' + ','.join(attr.enum)
                        ttl = 'NXDL attribute enum: '
                        ttl += '/'.join((nx_class_name, field_name))
                        ttl += '@' + k
                        self.new_finding(ttl, aname, f, m)
                elif attr.required:
                    if k not in ('name', 'minOccurs', 'maxOccurs',):
                        nm = 'NXDL attribute: ' + '/'.join((nx_class_name, field_name))
                        nm += '@' + k
                        m = 'required attribute not found'
                        self.new_finding(nm, dataset.name, finding.WARN, m)

        minO = defaults['minOccurs']
        maxO = defaults['maxOccurs']
        required_name = defaults['nameType'] == 'specified'
        if int(minO) > 0 and required_name:
            f = {True: finding.OK, False: finding.WARN}[target_exists]
            m = {True: '', False: ' not'}[target_exists] + ' found'
            nm = group.name + '/' + field_name
            self.new_finding(nx_class_name+' required field', nm, f, m)

            t = len(dataset.shape) == len(rules.dims)   # check rank against specification
            f = {True: finding.OK, False: finding.WARN}[t]  # TODO: ? change WARN to NOTE ?
            m = {True: 'matches', False: 'does not match'}[target_exists] + ' NXDL specification'
            self.new_finding(nx_class_name+' field rank', nm, f, m)

        if target_exists:
            if str(dataset.dtype).startswith('|S'):
                t = type('numpy string array') in nx_type
            else:
                t = dataset.dtype in nx_type
            f = {True: finding.OK, False: finding.WARN}[t]
            m = str(dataset.dtype) + {True: ' : ', False: ' not '}[t] + defaults['type']
            nm = group.name + '/' + field_name
            ttl = '/'.join((nx_class_name, field_name))
            self.new_finding('NXDL data type: '+ttl, nm, f, m)
        
        # TODO: #16 check if unknown names are allowed to be flexible 

    def validate_item_name(self, h5_addr, key=None):
        '''
        validate *h5_addr* using *validItemName* regular expression
        
        This is used for the names of groups, fields, links, and attributes.
        
        :param str h5_addr: full HDF5 address of item, for reference only,
            for attributes, use an @ symbol, such as these examples:
            
            =============================    ============
            *h5_addr*                        *short_name*
            =============================    ============
            ``/entry/user``                  ``user``
            ``/entry/data01/data``           ``data``
            ``/entry/data01/data@signal``    ``signal``
            =============================    ============

        :param str key: named key to search, default: None (``validItemName``)

        This method will separate out the last part of the name for validation.  
        Then, it is tested against the strict or relaxed regular expressions for 
        a valid item name.  The finding for each name is classified by the
        next table:
        
        =====  =======  =======  ================================================================
        order  finding  match    description
        =====  =======  =======  ================================================================
        1      OK       strict   matches most stringent NeXus specification
        2      NOTE     relaxed  matches NeXus specification that is most generally accepted
        3      ERROR    UTF8     specific to strings with UnicodeDecodeError (see issue #37)
        4      WARN     HDF5     acceptable to HDF5 but not NeXus
        =====  =======  =======  ================================================================
        
        :see: http://download.nexusformat.org/doc/html/datarules.html?highlight=regular%20expression
        '''
        if key is None:
            key_relaxed = 'validItemName'
            key_strict = 'validItemName-strict'

            # h5_addr = obj.name
            short_name = h5_addr.split('/')[-1].split('@')[-1]
            if short_name == 'NX_class':
                # special case
                self.new_finding('NeXus internal attribute', 
                                 h5_addr, 
                                 finding.OK, 
                                 'marks this HDF5 group as NeXus group')
                return
            
            # strict match: [a-z_][a-z\d_]*
            # flexible match: [A-Za-z_][\w_]*  but gets finding.WARN per manual
    
            p = self.patterns[key_strict]
            m = p.match(short_name)
            if m is not None and m.string == short_name:
                f = finding.OK
                key = key_strict
                msg =  'strict re: ' + p.regexp_pattern_str
            else:
                p = self.patterns[key_relaxed]
                m = p.match(short_name)
                if m is not None and m.string == short_name:
                    f = finding.NOTE
                    key = key_relaxed
                    msg =  'relaxed re: ' + p.regexp_pattern_str
                else:
                    # test if string rendering raises UnicodeDecodeError
                    key = 'validItemName'
                    msg = 'valid HDF5 item name, not valid with NeXus'
                    try:    # to raise the exception
                        _test = '%s' % str(m)
                        f = finding.WARN
                    except UnicodeDecodeError as _exc:
                        f = finding.ERROR
                        msg += ', UnicodeDecodeError'
        else:
            # TODO: validate h5_addr against other keys
            pass

        self.new_finding(key, h5_addr, f, msg)
    
    def validate_default_plot(self):
        '''
        check that data file defines the default plottable data
        
        :see: http://download.nexusformat.org/doc/html/datarules.html#find-the-plottable-data
        '''
        candidates = self.identify_default_plot_candidates()
        if self.default_plot_addr_v3(candidates['v3']) is not None:
            return
        elif self.default_plot_addr_v2(candidates['v2']) is not None:
            return
        elif self.default_plot_addr_v1(candidates['v1']) is not None:
            return
        
        m = 'no default plot: not a NeXus file'
        self.new_finding('NeXus default plot', '/NXentry/NXdata/field', finding.ERROR, m)
    
    def identify_default_plot_candidates(self):
        '''
        find the HDF5 addresses that might provide the default plottable data
        
        :see: http://download.nexusformat.org/doc/html/datarules.html#find-the-plottable-data
        :see: http://download.nexusformat.org/doc/html/preface.html?highlight=class%20path#class-path-specification
        
        There are different methods to identify the default data to be plotted.
        These can be distinguished by differences in the NeXus class path
        (the sequence of NeXus classes and other elements that describe an object in
        a NeXus HDF5 data file).  As used here, the text ``field`` is used
        instead of the name of the field (as shown in the NeXus manual) but the name of the
        attribute is given.
        
        ===========   =======================================
        version       NeXus classpath signature
        ===========   =======================================
        v3            /NXentry/NXdata@signal
        v3+niac2014   /@default/NXentry@default/NXdata@signal
        v2            /NXentry/NXdata/field@signal
        v1            /NXentry/NXdata/field@signal
        ===========   =======================================
        
        Versions *v1* and *v2* differ in their use of other attributes
        such as *axes* (v2) versus *axis* (v1) and *primary* (v1).
        with other attributes such as */NXentry/NXdata/field2@primary*.
        Since these other attributes are not always present, or
        might be used to indicate alternatives, a test for *v1*
        can fail due to both false negatives and false positives.
        '''
        # prepare dictionaries of candidates for the default plot
        candidates = dict(v1 = {}, v2 = {}, v3 = {})
        for node_name in self.h5:
            node = self.h5[node_name]
            if h5structure.isNeXusGroup(node, 'NXentry'):
                for subnode_name in node:
                    subnode = node[subnode_name]
                    if h5structure.isNeXusGroup(subnode, 'NXdata'):
                        if subnode.attrs.get('signal') is not None:
                            k = subnode.name + '@signal'
                            candidates['v3'][k] = '/NXentry/NXdata@signal'
                        for ss_node_name in subnode:
                            try:
                                ss_node = subnode[ss_node_name]
                            except KeyError:
                                continue
                            if not h5structure.isNeXusDataset(ss_node):
                                continue
                            if ss_node.attrs.get('signal') is not None:
                                k = ss_node.name + '@signal'
                                candidates['v2'][k] = '/NXentry/NXdata/field@signal'
                                candidates['v1'][k] = '/NXentry/NXdata/field@signal'
        return candidates

    def default_plot_addr_v1(self, group_dict):
        '''
        return the HDF5 address of the v1 default plottable data or None
        
        :see: http://download.nexusformat.org/doc/html/datarules.html#version-1
        '''
        default_plot_addr = []
        for primary_field_addr, nx_classpath in group_dict.items():
            title = 'NXdata group default plot v1'
            # need the NXdata group of this field
            primary = self.h5[primary_field_addr.split('@')[0]]
            nxdata_addr = '/'.join(primary.name.split('/')[:-1])
            nxdata = self.h5[nxdata_addr]
            signal_field_list = []
            for field_name in nxdata:
                field = nxdata[field_name]
                if h5structure.isNeXusDataset(field):
                    signal = self.get_hdf5_attribute(field, 'signal', report=True)
                    if signal is None:
                        continue
                    elif signal in (1, '1'):
                        signal_field_list.append(field)
                    else:
                        m = 'expected @signal=1, found: ' + signal
                        addr = field.name + '@signal'
                        self.new_finding(title, addr, finding.ERROR, m)
                        continue
                # TODO: @axis, @primary, and dimension scales
                # TODO: signal and dimension scales data shape

            if len(signal_field_list) == 1:
                m = 'NXdata group default plot using v1'
                self.new_finding(title, signal_field_list[0], finding.OK, m)
                default_plot_addr.append(signal_field_list[0])
            elif len(signal_field_list) == 0:
                m = 'NXdata group does not define a default plot using v1'
                self.new_finding(title, nxdata_addr, finding.WARN, m)
            else:
                m = 'NXdata group defines more than one default plot using v1'
                self.new_finding(title, nxdata_addr, finding.NOTE, m)
        
        cp = '/NXentry/NXdata/field@signal'
        title = 'NeXus default plot v1'
        if len(default_plot_addr) == 1:
            m = 'NeXus data file default plot defined'
            self.new_finding(title, default_plot_addr[0], finding.OK, m)
            return default_plot_addr[0]
        elif len(default_plot_addr) == 0:
            m = 'NeXus data file does not define a default plot using v1'
            #self.new_finding(title, cp, finding.WARN, m)
        else:
            m = 'NeXus data file defines more than one default plot using v1'
            self.new_finding(title, cp, finding.WARN, m)
            return default_plot_addr
    
    def default_plot_addr_v2(self, group_dict):
        '''
        return the HDF5 address of the v2 default plottable data or None
        
        :see: http://download.nexusformat.org/doc/html/datarules.html#version-2
        '''
        default_plot_addr = []
        for h5_addr, nx_classpath in group_dict.items():
            title = 'NeXus default plot v2'
            try:
                field = self.h5[h5_addr.split('@')[0]]
            except KeyError:
                continue
            signal = self.get_hdf5_attribute(field, 'signal', report=True)
            if signal in (1, '1'):
                m = nx_classpath + ' = 1'
                self.new_finding(title, field.name, finding.OK, m)
                default_plot_addr.append(field.name)
            else:
                m = 'expected @signal=1, found: ' + signal
                self.new_finding(title, h5_addr, finding.ERROR, m)
            # TODO: @axes and dimension scales    (see issue #41)
            # TODO: signal and dimension scales data shape

        cp = '/NXentry/NXdata/field@signal'
        title = 'NeXus default plot v2'
        if len(default_plot_addr) == 1:
            m = 'NeXus data file default plot defined using v2'
            self.new_finding(title, default_plot_addr[0], finding.OK, m)
            return default_plot_addr[0]
        elif len(default_plot_addr) == 0:
            m = 'NeXus data file does not define a default plot using v2'
            #self.new_finding(title, cp, finding.WARN, m)
        else:
            m = 'NeXus data file defines more than one default plot using v2'
            self.new_finding(title, cp, finding.NOTE, m)
            return default_plot_addr
    
    def default_plot_addr_v3(self, group_dict):
        '''
        return the HDF5 address of the v3 default plottable data or None
        
        :see: http://download.nexusformat.org/doc/html/datarules.html#version-3
        '''
        default_plot_addr = []
        for h5_addr, nx_classpath in group_dict.items():
            dimension_scales = []
            dimension_scales_ok = True  # assume until proven otherwise
            title = 'NXdata group default plot v3'
            nxdata = self.h5[h5_addr.split('@')[0]]
            signal_name = self.get_hdf5_attribute(nxdata, 'signal', report=True)
            if signal_name in nxdata:
                signal_data = nxdata[signal_name]
                m = 'NXdata@signal = ' + signal_name
                addr = nxdata.name
                self.new_finding(title, addr+'@signal', finding.OK, m)

                axes_names = self.get_hdf5_attribute(nxdata, 'axes', report=True)
                if axes_names is None:  # no axes attribute: use array indices as dimension scales
                    for dim in signal_data.shape:
                        dimension_scales.append(numpy.ndarray(dim))
                else:
                    if isinstance(axes_names, str):
                        for delim in (':', ' '):
                            # replace alternate delimiters (":", " ") with ","
                            axes_names = axes_names.replace(delim, ',')
                        axes_names = axes_names.split(',')
                    for axis_name in axes_names:
                        ttl = 'NXdata@axes'
                        if axis_name == '.':
                            pass
                        elif axis_name in nxdata:     # does axis exist?
                            m = 'axes dataset found: ' + axis_name
                            f = finding.OK
                            self.new_finding(ttl+'='+axis_name, addr+'@axes', f, m)
                            # check @AXISNAME_indices holds index of dimension scale data to use
                            # dimension scale = index 'indices' of nxdata[axis_name]
                            axis_data = nxdata[axis_name]
                            indices = self.get_hdf5_attribute(nxdata, axis_name+'_indices')
                            if indices is None:
                                if len(axis_data.shape) == 1:
                                    m = 'not provided, assume = 0'
                                    self.new_finding('NXdata@'+axis_name+'_indices', 
                                                     nxdata.name+'@'+axis_name+'_indices', 
                                                     finding.NOTE, 
                                                     m)
                                else:
                                    m = 'not provided, uncertain how to use'
                                    self.new_finding('NXdata@'+axis_name+'_indices', 
                                                     nxdata.name+'@'+axis_name+'_indices', 
                                                     finding.WARN, 
                                                     m)
                                    dimension_scales_ok = False
                            else:
                                if isinstance(indices, tuple):
                                    indices = numpy.ndarray(indices)
                                if isinstance(indices, numpy.ndarray) and len(indices) > 1:
                                    t = numpy.all(indices < len(signal_data.shape))
                                else:
                                    t = indices < len(signal_data.shape)
                                f = finding.TF_RESULT[t]
                                m = 'value = ' + str(indices)
                                m += {True:': ok', False:': invalid'}[t]
                                self.new_finding('NXdata@'+axis_name+'_indices', 
                                                     nxdata.name+'@'+axis_name+'_indices', 
                                                     f, 
                                                     m)
                                if not t:
                                    dimension_scales_ok = False

                            if len(axis_data.shape) == 1:
                                dimension_scales.append(axis_data)
                            elif len(axis_data.shape) == 2:
                                if not isinstance(indices, numpy.ndarray):
                                    dimension_scales.append(axis_data[indices])
                                else:
                                    for indx in indices:
                                        dimension_scales.append(axis_data[indx])
                            else:
                                m = axis_data.name + '@axes, axis=' + axis_name
                                m += ' has rank=' + str(len(axis_data.shape))
                                m += '\n This needs special handling.  Send email to the developer.'
                                raise ValueError(m)
                        else:
                            m = 'axes dataset not found: ' + axis_name
                            f = finding.WARN
                            self.new_finding(ttl+'='+axis_name, addr+'@axes', f, m)
                            dimension_scales_ok = False

                if len(dimension_scales) == len(signal_data.shape):
                    if len(dimension_scales) == 1:
                        length_ok = dimension_scales[0].shape[0] - signal_data.shape[0] in (0, 1)
                        # 0 : dimension scale values are bin centers
                        # 1 : dimension scale values are bin edges
                        if not length_ok:
                            ttl = 'dimension scale for NXdata@signal'
                            m = 'array lengths are not the same'
                            self.new_finding(ttl, signal_data.name, finding.WARN, m)
                            dimension_scales_ok = False
                    else:
                        for i, dscale in enumerate(dimension_scales):
                            length_ok = dscale.shape[0] - signal_data.shape[i] in (0, 1)
                            if not length_ok:
                                ttl = 'dimension scale for NXdata@signal[' + str(i) + ']'
                                m = 'array lengths are not the same'
                                self.new_finding(ttl, signal_data.name, finding.WARN, m)
                                dimension_scales_ok = False
                else:
                    m = 'rank(' + signal_name + ') != number of dimension scales'
                    self.new_finding('NXdata@signal rank', signal_data.name, finding.WARN, m)
                default_plot_addr.append(addr)
                if dimension_scales_ok:
                    m = 'dimension scale(s) verified'
                    self.new_finding('NXdata dimension scale(s)', nxdata.name, finding.OK, m)
            else:
                m = 'signal field not found: ' + signal_name
                self.new_finding(title, nx_classpath, finding.ERROR, m)
                continue

        cp = 'NXdata@signal'
        title = 'NeXus default plot v3'
        if len(default_plot_addr) == 1:
            m = 'NeXus data file default plot: /NXentry/NXdata@signal'
            self.new_finding(title, default_plot_addr[0], finding.OK, m)
            return default_plot_addr[0]
        elif len(default_plot_addr) == 0:
            m = 'NeXus data file does not define a default plot using v3'
            # self.new_finding(title, cp, finding.WARN, m)
        else:
            # use NIAC2014 terms to find unique address
            unique_list = self.default_plot_addr_v3_niac2014(default_plot_addr)
            if len(unique_list) == 1:
                m = 'NeXus data file default plot defined using v3+niac2014'
                self.new_finding(title, unique_list[0], finding.OK, m)
                return unique_list[0]
            else:
                m = 'NeXus data file defines more than one default plot'
                self.new_finding(title, cp, finding.NOTE, m)
                return default_plot_addr
    
    def default_plot_addr_v3_niac2014(self, address_list):
        '''
        return a list of default plottable data as directed by @default attributes
        
        :param [str] address_list: list of absolute HDF5 addresses with v3 default plottable data
        
        Each address fits this NeXus class path: /NXentry/NXdata/field  
        '''
        unique_list = []
        for k in address_list:
            nxentry_name = k.split('/')[1]
            root_default = self.get_hdf5_attribute(self.h5, 'default', nxentry_name)
            if root_default == nxentry_name:
                nxentry = self.h5[nxentry_name]
                nxdata_name = k.split('/')[2]
                nxentry_default = self.get_hdf5_attribute(nxentry, 'default', nxdata_name)
                if nxentry_default == nxdata_name:
                    unique_list.append(k)
        return unique_list
   
    def validate_numerical_dataset(self, dataset, group):
        '''
        review the units attribute of an HDF5 dataset

        :param obj dataset: instance of h5py.Dataset
        :param obj group: instance of h5py.Group or h5py.File, needed to check against NXDL
        '''
        if dataset.dtype not in self.data_types['NX_NUMBER']:
            return

        # check the units of numerical fields
        title = 'field@units'
        units = self.get_hdf5_attribute(dataset, 'units', report=True)
        t = units is not None
        f = {True: finding.OK, False: finding.NOTE}[t]
        msg = {True: 'exists', False: 'does not exist'}[t]
        if t:
            t = len(units) > 0
            f = {True: finding.OK, False: finding.NOTE}[t]
            msg = {True: 'value: ' + units, False: 'has no value'}[t]
        self.new_finding(title, dataset.name + '@units', f, msg)
        # TODO: compare value of dataset@units with NXDL@units specification
        # this could easily require a deep analysis
        
        # TODO: issue #13: check field dimensions against "rank" : len(shape) == len(NXDL/dims) 
        shape = dataset.shape
        if shape != (1,):   # ignore scalars
            __ = None   # used as a NOP breakpoint after previous definition

    def get_hdf5_attribute(self, obj, attribute, default=None, report=False):
        '''
        HDF5 attribute strings might be coded in several ways
        
        :param obj obj: instance of h5py.File, h5py.Group, or h5py.Dataset
        :param str attribute: name of requested attribute
        :param obj default: value if attribute not found (usually str)
        :param bool report: check & report if value is a variable length string (actually an ndarray)
        '''
        a = obj.attrs.get(attribute, default)
        if isinstance(a, numpy.ndarray):
            if report:
                gname = obj.name + '@' + attribute
                if len(a) > 1:
                    msg = 'variable length string array: ' + str(a)
                else:
                    msg = 'variable length string: ' + str(a)
                self.new_finding('attribute data type', gname, finding.NOTE, msg)
            if len(a) == 1:
                a = a[0]
        return a

    def reconstruct_classpath(self, h5_address, *args, **kwargs):
        '''
        build the classpath from the h5_address
        '''
        path = h5_address.lstrip('/').split('@')[0]
        if len(path) == 0:
            return
 
        # reconstruct the NeXus classpath
        cp = ''     # classpath to be built
        hp = ''     # HDF5 address to be built
        for item in path.split('/'):
            hp += '/' + item
            if hp in self.h5:
                try:
                    item = self.h5[hp]
                except KeyError as _exc:
                    cp += '/missing_external_file_link'
                    continue
                if h5structure.isHdf5Dataset(item):
                    cp += '/field'
                else:
                    obj = self.h5[hp]
                    nx_class = self.get_hdf5_attribute(obj, 'NX_class', '-')
                    cp += '/' + nx_class
        if '@' in h5_address:
            cp += '@' + h5_address.split('@')[-1]
         
        return cp

    def new_finding(self, test_name, h5_address, status, comment):
        '''
        accumulate a list of findings
        
        :param str test_name: brief name of this test
        :param str h5_address: HDF5 address
        :param obj status: instance of finding.ValidationResultStatus,
                should be the same text as other instances of this test
        :param str comment: free-form explanation
        '''
        addr = str(h5_address)
        unique_key = addr + ':' + test_name
        if unique_key in self.__unique_findings__:
            # ensure that each test is only recorded once
            return
        f = finding.Finding(test_name, addr, status, comment)
        self.findings.append(f)
        self.__unique_findings__[unique_key] = f
        if addr not in self.addresses:
            # accumulate a dictionary of HDF5 object addresses
            self.addresses[addr] = finding.CheckupResults(addr)
            self.addresses[addr].classpath = self.reconstruct_classpath(addr)
        self.addresses[addr].findings.append(f)
    
    def report_findings(self, statuses=()):
        '''
        make a table of the validation findings
        
        :param statuses: List (or tuple) of finding statuses to be shown.
            Several lists have been pre-defined for convenience:

            :data:`finding.SHOW_ALL`        ``(OK, NOTE, WARN, ERROR, TODO, UNUSED, COMMENT)``
            :data:`finding.SHOW_NOT_OK`     ``(ERROR, WARN)``
            :data:`finding.SHOW_ERRORS`     ``(WARN, ERROR, TODO, UNUSED)``
            
            See :mod:`finding` for details.

        '''
        import pyRestTable

        t = pyRestTable.Table()
        t.labels = 'address validation status comment(s)'.split()
        for f in self.findings:
            if f.status in statuses:
                t.rows.append((f.h5_address, f.test_name, f.status, f.comment))
        return t.reST()
    
    def report_findings_summary(self):
        '''
        make a summary table of the validation findings (count how many of each status)
        '''
        import pyRestTable

        # count each category
        summary = collections.OrderedDict()
        for k in finding.VALID_STATUS_LIST:
            summary[str(k.key)] = 0
        xref = {str(k): k for k in finding.VALID_STATUS_LIST}
        for f in self.findings:
            summary[str(f.status)] += 1

        t = pyRestTable.Table()
        t.labels = 'status count description'.split()
        for k, v in summary.items():
                t.rows.append((k, v, xref[k].description))
        t.rows.append(('--', '--', '--'))
        t.rows.append(('TOTAL', len(self.findings), '--'))
        return t.reST()
    
    def report_classpath(self):
        '''
        make a table of the known NeXus class paths
        '''
        import pyRestTable
        t = pyRestTable.Table()
        t.labels = 'HDF5-address  NeXus-classpath'.split()
        for k, v in self.addresses.items():
            t.rows.append((k, v.classpath))
        return t.reST()
    
    def missing_file_link(self, text):
        '''
        Return file name if error message from KeyError due to missing external file link, else None
        
        Such as::
        
            Unable to open object (Unable to open file: name = 'data\\../nt15698-1/processing/waxs_mask.nxs', errno = 2, error message = 'no such file or directory', flags = 0, o_flags = 0)
        
        Returns::
        
            data\\../nt15698-1/processing/waxs_mask.nxs
        
        '''
        filename = None
        m1 = 'Unable to open object (Unable to open file: name = '
        p1 = text.find(m1)
        if p1 >= 0:
            p2 = text.find(', errno =')
            filename = text[p1 + len(m1) : p2].strip("'")
        return filename

if __name__ == '__main__':
    print("Start this module using:  python main.py validate ...")
    exit(0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------

'''
developer: exercise various aspects of the main user interface file
'''

import sys


def demo_prep():
    ''' '''
    sys.argv.append('de')
    # sys.argv.append('-h')
    logfile_prep()


def update_prep():
    ''' '''
    sys.argv.append('up')
    # sys.argv.append('-h')
    sys.argv.append('-f')
    logfile_prep()


def structure_prep():
    ''' '''
    sys.argv.append('st')
    # sys.argv.append('-h')
    logfile_prep()
    datafile_prep()


def validate_prep():
    ''' '''
    sys.argv.append('va')
    # sys.argv.append('-h')
    logfile_prep()
    # sys.argv.append('--report')
    # sys.argv.append('ERROR,WARN')
    # datafile_prep()
    sys.argv.append('data/i22-331015.nxs')


def logfile_prep():
    ''' '''
    # sys.argv.append('--logfile')
    # sys.argv.append('_punx_.log')
    # sys.argv.append('--interest')
    # sys.argv.append('1')


def datafile_prep():
    ''' '''
    # sys.argv.append('data/writer_1_3.hdf5')
    # sys.argv.append('data/writer_2_1.hdf5')
    # sys.argv.append('data/chopper.nxs')
    # sys.argv.append('data/compression.h5')
    # sys.argv.append('data/prj_test.nexus.hdf5')
    # sys.argv.append('data/scan101.nxs')
    # sys.argv.append('data/has_default_attribute_error/external_master.hdf5')
    # sys.argv.append('data/has_default_attribute_error/external_angles.hdf5')
    # sys.argv.append('data/verysimple.nx5')
    # sys.argv.append('data/1998spheres.h5')
    # sys.argv.append('data/cs_af1410.h5')
    # sys.argv.append('data/USAXS_flyScan_GC_M4_NewD_15.h5')
    sys.argv.append('data/33id_spec_22_2D.hdf5')


if __name__ == '__main__':
    import __init__
    import main
    __init__.DEFAULT_LOG_LEVEL = __init__.NOISY
    # sys.argv.append('-h')
    
    # demo_prep()
    # update_prep()
    # structure_prep()
    validate_prep()

    main.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

'''
Load and/or document the structure of a NeXus NXDL class specification

.. autosummary::
   
   ~get_nxdl_rules
   ~get_ns_dict
   ~validate_NXDL
   ~NX_mixin
   ~NX_definition
   ~NX_attribute
   ~NX_field
   ~NX_group
   ~NX_link
   ~NX_symbols
   ~get_NXDL_specifications

* :class:`NX_definition`: the structure
* define a text renderer method for that class

.. rubric:: about attributes

There are three types of attributes this code must handle:

1. XML attributes
   
   These are metadata applied to various elements in an XML file.  Example::
   
       <xs:element minOccurs="0" ...
    
    XML attributes can be stored in a key:value dictionary where value is a string.
   
2. NXDL attributes
   
   These are attribute elements presented in an NXDL file.  Example::
   
       <nx:attribute ...
    
    NXDL attributes can be stored in a key:object dictionary where object is an instance of NX_attribute.
    Content of that object is a dictionary of XML attributes and various objects which define the
    structure of the defined NXDL attribute.

3. HDF5 data file attributes
   
   These are metadata applied to various groups or fields in an HDF5 data file.  Example::
   
       /entry@default=...
    
    HDF5 data file attributes can be stored in a key:value dictionary where value is a string.

'''

__url__ = 'http://punx.readthedocs.org/en/latest/nxdlstructure.html'



import collections
import lxml.etree
import os

import __init__
import cache
import nxdl_rules
import validate


PROGRAM_NAME = 'nxdlstructure'
INDENTATION_UNIT = '  '
listing_category = None

__singleton_nxdl_rules__ = None


def get_nxdl_rules():
    '''
    parse and cache the XML Schema file (nxdlTypes.xsd) as an XML document only once
    '''
    global __singleton_nxdl_rules__

    if __singleton_nxdl_rules__ is None:
        __singleton_nxdl_rules__ = nxdl_rules.NxdlRules()

    return __singleton_nxdl_rules__


def get_ns_dict():
    '''
    get the dictionary of XML namespace substitutions
    '''
    return get_nxdl_rules().nxdl.ns


def validate_NXDL(nxdl_file_name):
    '''
    Validate a NeXus NXDL file
    '''
    validate.validate_xml(nxdl_file_name)


class NX_mixin(object):
    '''
    common components available to all subclasses
    
    :param obj node: XML object
    
    .. autosummary::
    
        ~get_NX_type
        ~get_NX_units
        ~get_element_data
        ~add_object
        ~render_group
    
    '''
    element = 'mixin - must override in subclass'
    
    def __init__(self, node):
        node_name = node.get('name') or node.get('type')
        if node_name is not None:
            if node_name.startswith('NX'):
                node_name = node_name[2:]
            self.name = node_name

        # CAREFUL:
        # there are several kinds of attributes
        # 1. XML attributes of XML elements defining the NXDL structure
        # 2. NXDL attribute ... make this clear ...  THESE are in self.attributes['NXDL.xml']
        # 3. custom XML Schema attributes
        # The default attribute values are set from the list above, in that order


        self.attributes = {}
        self.attributes['nxdl.xsd'] = {}    # attributes defined in nxdl.xsd XML Schema
        self.attributes['NXDL.xml'] = {}    # attributes defined in NXDL.xml files
        self.attributes['defined'] = {}     # XML attributes defined in this instance of the element
        self.attributes['defaults'] = {}    # combination of the above

        if self.element == 'definition':
            defaults = get_nxdl_rules().nxdl
        else:
            defaults = get_nxdl_rules().nxdl.children[self.element]
        self.attributes['nxdl.xsd'] = {k: v for k, v in defaults.attrs.items()}

    def __str__(self, *args, **kwargs):
        '''
        canonical string representation of this object
        '''
        try:
            return self.name
        except:
            return object.__str__(self, *args, **kwargs)
    
    def get_NX_type(self, node):
        '''
        return the NeXus type of the XML node
        '''
        return node.get('type', 'NX_CHAR')
    
    def get_NX_units(self, node):
        '''
        return the units attribute of the XML node
        '''
        return node.get('units', '')
    
    def get_element_data(self, node, category):
        '''
        parse the elements defined within this XML node 
        '''
        self.fields = {}
        self.groups = {}
        self.links = {}
        self.symbols = {}
        
        switch_block = dict(
            attribute = [NX_attribute, self.attributes['NXDL.xml']],
            field = [NX_field, self.fields],
            group = [NX_group, self.groups],
            link = [NX_link, self.links],
            symbols = [NX_symbols, self.symbols],
        )
        for subnode in node:
            if isinstance(subnode.tag, str):    # do not process XML Comments
                tag = subnode.tag.split('}')[-1]
                if tag in switch_block:         # only handle certain elements
                    NX_handler, nx_dict = switch_block[tag]
                    obj = NX_handler(subnode, category)
                    self.add_object(nx_dict, obj)

    def add_object(self, db, obj):
        '''
        add object ``obj`` to database ``db`` (dictionary)
        '''
        try:
            name = obj.name
        except AttributeError as _exc:
            name = 'symbol'    # FIXME: NX_symbols.name is not defined
        if name in db:
            for i in range(1, 1+len(db)):
                nm = name + '_' + str(i)
                if nm not in db: 
                    name = nm
                    break
        db[name] = obj

    def render_group(self, group):
        '''
        return a string with text representation of this ``group``
        '''
        indentation = ' '*2
        t = []
        for nx_dict in (group.attributes['NXDL.xml'], group.fields, group.links):
            for _k, v in sorted(nx_dict.items()):
                t.append(str(v))
        for _k, v in sorted(group.groups.items()):
            t.append(str(v))
            for line in group.render_group(v).splitlines():
                t.append(indentation + line)
        return '\n'.join(t)


class NX_definition(NX_mixin):
    '''
    Contains the complete structure of a NXDL definition, without documentation
    
    :param str nxdl_file: name of file with NXDL definition (ends with ``.nxdl.xml``)
    '''
    element = 'definition'
    
    def __init__(self, nxdl_file):
        self.nxdl_file_name = nxdl_file
        if not os.path.exists(nxdl_file):
            raise IOError('file does not exist: ' + nxdl_file)
        validate_NXDL(nxdl_file)
        
        # parse the XML content now
        tree = lxml.etree.parse(self.nxdl_file_name)
        root = tree.getroot()

        NX_mixin.__init__(self, root)
        self.title = root.get('name')
        self.category = root.attrib["category"]
        
        for subnode in root:
            if isinstance(subnode.tag, str):    # do not process XML Comments
                tag = subnode.tag.split('}')[-1]
                if tag == 'attribute':
                    obj = NX_attribute(subnode, self.category)
                    self.add_object(self.attributes['NXDL.xml'], obj)

        # get the attributes specified in THIS group element
        for k, v in root.attrib.items():
            if k not in ('name',):
                self.attributes['defined'][k] = v

        default_attributes = self.attributes['defaults']
        for k, v in self.attributes['nxdl.xsd'].items():
            default_attributes[k] = v.default_value
        for k, v in self.attributes['NXDL.xml'].items():
            default_attributes[k] = None        # means: undefined
        for k, v in self.attributes['defined'].items():
            default_attributes[k] = v

        self.get_element_data(root, self.category)
    
    def __str__(self):
        return self.title + ' : ' + self.category
    
    def render(self):
        '''
        '''
        indentation = ' '*2
        t = [self.title,]
        for line in self.render_group(self).splitlines():
            t.append(indentation + line)
        return '\n'.join(t)

    def getSubGroup_NX_class_list(self):
        '''
        list the groups used in this NXDL specification by NX_class name
        ''' 
        return sorted({v.NX_class: None for v in self.groups.values()}.keys())


class NX_attribute(NX_mixin):
    '''
    NXDL attribute
    '''
    element = 'attribute'

    def __init__(self, node, category):
        NX_mixin.__init__(self, node)

        # get the attributes specified in THIS field element
        for k, v in node.attrib.items():
            if k not in ('name',):
                self.attributes['defined'][k] = v
        
        default_attributes = self.attributes['defaults']
        for k, v in self.attributes['nxdl.xsd'].items():
            default_attributes[k] = v.default_value
        for k, v in self.attributes['NXDL.xml'].items():
            default_attributes[k] = None        # means: undefined
        for k, v in self.attributes['defined'].items():
            default_attributes[k] = v

        self.enum = []
        for n in node.xpath('nx:enumeration/nx:item', namespaces=get_ns_dict()):
            self.enum.append( n.attrib['value'] )

    def __str__(self, *args, **kwargs):
        try:
            s = '@' + self.name
            s += ' : ' + self.attributes['defaults']['type']
            if len(self.enum):
                s += ' = '
                s += ' | '.join(self.enum)
            return s
        except:
            return NX_mixin.__str__(self, *args, **kwargs)


class NX_field(NX_mixin):
    '''
    NXDL field
    '''
    element = 'field'

    def __init__(self, node, category):
        NX_mixin.__init__(self, node)
        
        self.dims = self.field_dimensions(node)

        self.enum = []
        for n in node.xpath('nx:enumeration/nx:item', namespaces=get_ns_dict()):
            self.enum.append( n.attrib['value'] )

        # walk through all <attribute /> elements declared in the NXDL
        # use same code as above
        switch_block = dict(
            attribute = [NX_attribute, self.attributes['NXDL.xml']],
            #field = [NX_field, self.fields],
            #group = [NX_group, self.groups],
            #link = [NX_link, self.links],
            #symbols = [NX_symbols, self.symbols],
            #dimensions = ?
        )

        for subnode in node:
            if isinstance(subnode.tag, str):    # do not process XML Comments
                tag = subnode.tag.split('}')[-1]
                if tag in switch_block:         # only handle certain elements
                    NX_handler, nx_dict = switch_block[tag]
                    obj = NX_handler(subnode, category)
                    self.add_object(nx_dict, obj)

        # get the attributes specified in THIS field element
        for k, v in node.attrib.items():
            if k not in ('name',):
                self.attributes['defined'][k] = v

        default_attributes = self.attributes['defaults']
        for k, v in self.attributes['nxdl.xsd'].items():
            default_attributes[k] = v.default_value
        if category in ('application',):
            # this rule is not in the XML Schema: issue #266
            # fields are required in application definitions
            default_attributes['minOccurs'] = 1
        for k, v in self.attributes['NXDL.xml'].items():
            default_attributes[k] = None        # means: undefined
        for k, v in self.attributes['defined'].items():
            default_attributes[k] = v

    def __str__(self):
        s = self.name
        if 'type' in self.attributes['defined']:
            s += ' : ' + self.attributes['defined']['type']
        if len(self.dims):
            s += '[' + ', '.join(map(str, self.dims)) + ']'
        if len(self.enum):
            s += ' = ' + ' | '.join(self.enum)
        return s
    
    def field_dimensions(self, parent):
        '''
        '''
        node_list = parent.xpath('nx:dimensions', namespaces=get_ns_dict())
        if len(node_list) != 1:
            return []

        dims = {}
        for subnode in node_list[0].xpath('nx:dim', namespaces=get_ns_dict()):
            index = int(subnode.get('index'))
            value = subnode.get('value')
            if not value:
                value = 'ref(%s)' % subnode.get('ref')
            if index == 0:
                # index="0": cannot know to which dimension this applies a priori
                value = '*' + str(value) + '*'
            dims[index] = value
        
        if len(dims) == 0:
            # read the "rank" attribute and create array of that length with '*'
            rank = int(node_list[0].get('rank'))
            return ['*' for _ in range(rank)]
        if len(dims) == 1 and min(dims.keys()) == 0:
            '''
                The rank of the ``data`` must satisfy
                ``1 <= dataRank <= NX_MAXRANK=32``.  
                At least one ``dim`` must have length ``n``.
            '''
            pass    # index="0": cannot know to which dimension this applies a priori
        elif min(dims.keys()) != 1 or max(dims.keys()) != len(dims):
            msg = 'dimensions not specified properly: ' + str(dims)
            raise KeyError(msg)
        return [dims[k] for k in sorted(map(int, dims.keys()))]


class NX_group(NX_mixin):
    '''
    NXDL group
    '''
    element = 'group'

    def __init__(self, node, category):
        NX_mixin.__init__(self, node)
        
        self.NX_class = node.get('type', None)
        if self.NX_class is None:
            msg = 'group has no type, this is an error, name = ' + self.name
            raise ValueError(msg)

        # get the attributes specified in THIS group element
        for k, v in node.attrib.items():
            if k not in ('name',):
                self.attributes['defined'][k] = v

        default_attributes = self.attributes['defaults']
        for k, v in self.attributes['nxdl.xsd'].items():
            default_attributes[k] = v.default_value
        if category in ('application',):
            # this rule is not in the XML Schema: issue #266
            # fields are required in application definitions
            default_attributes['minOccurs'] = 1
        for k, v in self.attributes['NXDL.xml'].items():
            default_attributes[k] = None        # means: undefined
        for k, v in self.attributes['defined'].items():
            default_attributes[k] = v

        self.get_element_data(node, category)

    def __str__(self):
        s = self.name
        s += ' : ' + self.NX_class
        return s


class NX_link(NX_mixin):
    '''
    NXDL link
    '''
    element = 'link'

    def __init__(self, node, category):
        NX_mixin.__init__(self, node)
        
        self.target = node.get('target')
    
    def __str__(self):
        return self.name + ' --> ' + self.target


class NX_symbols(NX_mixin):
    '''
    NXDL symbols table
    '''

    element = 'symbols'

    def __init__(self, node, category):
        NX_mixin.__init__(self, node)

        # TODO: finish this


def _get_specs_from_pickle_file():
    '''
    try to get the NXDL dictionary from the pickle file
    
    :return: dict with definitions or None
    '''
    qset = cache.qsettings()
    pfile = cache.get_pickle_file_name(qset.cache_dir())
    if not os.path.exists(pfile):
        return

    sha = qset.getKey('git_sha')
    try:
        nxdl_dict = cache.read_pickle_file(pfile, sha)
    except (AttributeError, ImportError):
        # could not read the pickle file, suggest user update the cache
        nxdl_dict = None
    if nxdl_dict is None:
        return

    return  nxdl_dict


def _get_specs_from_NXDL_files():
    '''
    get the NXDL dictionary from the NXDL files in the cache
    
    :return: dict with definitions
    '''
    basedir = cache.get_nxdl_dir()

    path_list = [
        os.path.join(basedir, 'base_classes'),
        os.path.join(basedir, 'applications'),
        os.path.join(basedir, 'contributed_definitions'),
    ]
    nxdl_file_list = []
    for path in path_list:
        if not os.path.exists(path):
            raise IOError('no definition available, cannot find ' + path)
        for fname in sorted(os.listdir(path)):
            if fname.endswith('.nxdl.xml'):
                nxdl_file_list.append(os.path.join(path, fname))
    
    nxdl_dict = collections.OrderedDict()
    for nxdl_file_name in nxdl_file_list:
        obj = NX_definition(nxdl_file_name)
        nxdl_dict[obj.title] = obj
    
    return nxdl_dict


def get_NXDL_specifications():
    '''
    return a dictionary of NXDL structures, keyed by NX_class name
    '''
    return _get_specs_from_pickle_file() or _get_specs_from_NXDL_files()


def _developer():
    'working on issue #4'
    import logs
    __init__.LOG_MESSAGE = logs.to_console
    nxdl_dict = _get_specs_from_NXDL_files()
    print(sorted(nxdl_dict.keys()))


if __name__ == '__main__':
    # TODO: remove for production
    # print("Start this module using:  python main.py structure ...")
    # exit(0)
    _developer()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

'''
Perform various analyses on the NXDL files

1. load ALL the NXDL classes and prepare directed graph of the base class inheritance.

.. autosummary::
   
   ~ab_counter
   ~ab_groups
   ~base_class_hierarchy

.. warning:: This module is under development, not certain to work properly
'''


import nxdlstructure


ab = {}
def ab_counter(a, b):
    '''
    count instances of 'b' is a child of 'a'
    '''
    if a not in ab:
        ab[a] = {}
    if b not in ab[a]:
        ab[a][b] = 0
    ab[a][b] += 1


def ab_groups(p, g):
    '''
    recurse base class hierarchy and count parent:child instances
    '''
    for sub in g.groups.values():
        if nxdl_dict[p].category in ('base class',):
            ab_counter(p, sub.NX_class)
        ab_groups(sub.NX_class, sub)


def base_class_hierarchy(nxdl_dict):
    '''
    print commands to generate a directed graph (for graphviz) of the base class hierarchy
    '''
    # print len(nxdl_dict)
    for v in nxdl_dict.values():
        ab_groups(v.title, v)
    
    # print out commands for graphviz
    #   dot -Tpng graph.dot -o graph.png
    print('digraph G {')
    print('  rankdir=LR;')
    print('  NXentry [shape=box,color=red];')
    print('  NXdata [shape=box,color=red];')
    for p, db in sorted(ab.items()):
        for c, n in sorted(db.items()):
            if p == 'NXentry' and c == 'NXdata':
                print('  %s -> %s [weight=%d,shape=box,style=filled,color=red];' % (p, c, n+100))
            else:
                print('  %s -> %s [weight=%d];' % (p, c, n))
    print('}')


if __name__ == '__main__':
    nxdl_dict = nxdlstructure.get_NXDL_specifications()
    base_class_hierarchy(nxdl_dict)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

'''
print the structure of an HDF5 file

.. autosummary::
    
    ~h5structure
    ~isHdf5File
    ~isHdf5Group
    ~isHdf5Dataset
    ~isHdf5Link
    ~isHdf5ExternalLink
    ~isNeXusFile
    ~isNeXusFile_ByNXdataAttrs
    ~isNeXusFile_ByAxes
    ~isNeXusFile_ByAxisAttr
    ~isNeXusGroup
    ~isNeXusDataset
    ~isNeXusLink

'''


__url__ = 'http://punx.readthedocs.org/en/latest/h5structure.html'

import os       #@UnusedImport
import sys      #@UnusedImport
import h5py
import numpy

import __init__


class h5structure(object):
    '''
    show structure of any HDF5 data file
    
    Example usage showing default display::
    
        mc = h5structure(filename)
        mc.array_items_shown = 5
        show_attributes = False
        txt = mc.report(show_attributes)
    
    '''
    requested_filename = None
    isNeXus = False
    array_items_shown = 5

    def __init__(self, filename):
        '''store filename and test if file is NeXus HDF5'''
        self.requested_filename = filename
        self.filename = None
        self.show_attributes = True
        if os.path.exists(filename):
            self.filename = filename
            self.isNeXus = isNeXusFile(filename)
        else:
            raise __init__.FileNotFound(filename)

    def report(self, show_attributes=True):
        '''
        return the structure of the HDF5 file in a list of strings
        
        The work of parsing the datafile is done in this method.
        '''
        if self.filename is None: return None
        if not os.path.exists(self.filename):
            raise __init__.FileNotFound(self.filename)
        self.show_attributes = show_attributes
        try:
            f = h5py.File(self.filename, 'r')
        except IOError:
            raise __init__.HDF5_Open_Error(self.filename)
        txt = self.filename
        if self.isNeXus:
            txt += " : NeXus data file"
        structure = self._renderGroup(f, txt, indentation = "")
        f.close()
        return structure

    def _renderGroup(self, obj, name, indentation = "  "):
        '''return a [formatted_string] with the contents of the group'''
        s = []
        nxclass = obj.attrs.get('NX_class', '')
        if len(nxclass) > 0:
            if isinstance(nxclass, numpy.ndarray):      # attribute reported as DATATYPE SIMPLE
                nxclass = nxclass[0]                    # convert as if DATATYPE SCALAR
            nxclass = ":" + nxclass
        s += [ indentation + name + nxclass ]
        s += self._renderAttributes(obj, indentation)
        # show datasets and links next
        groups = []
        for itemname in sorted(obj):
            linkref  = obj.get(itemname, getlink=True)
            try:
                # this will fail for external links if file is not available
                classref = obj.get(itemname, getclass=True)
            except KeyError:
                classref = None

            if classref is None:
                s += [ '%s  %s: external file missing' % (indentation, itemname) ]
                fmt = '%s    %s = %s'
                s += [ fmt % (indentation, '@file', linkref.filename) ]
                s += [ fmt % (indentation, '@path', linkref.path) ]
            else:
                value = obj.get(itemname)
                if isNeXusLink(value):
                    s += self._renderLinkedObject(value, itemname, indentation+"  ")
                elif isHdf5Group(value) or isHdf5File(value):
                    groups.append(value)
                    # TODO: issue #18: report external group links in the right place
                    # The problem is the link file and path need to be fed into the
                    # next call to _renderGroup().  No such design exists now for that. 
                elif isHdf5Dataset(value):
                    s += self._renderDataset(value, itemname, indentation+"  ")
                    if isHdf5ExternalLink(linkref):
                        # When "classref" is defined, then external data is available
                        fmt = '%s    %s = %s'
                        s += [ fmt % (indentation, '@file', linkref.filename) ]
                        s += [ fmt % (indentation, '@path', linkref.path) ]
                else:
                    msg = "unidentified %s: %s, %s", itemname, repr(classref), repr(linkref)
                    raise Exception(msg)

        for value in groups:        # show things that look like groups
            itemname = value.name.split("/")[-1]
            s += self._renderGroup(value, itemname, indentation+"  ")
        
        return s

    def _renderAttributes(self, obj, indentation = "  "):
        '''return a [formatted_string] with any attributes'''
        s = []
        if self.show_attributes:
            for name in obj.attrs:
                try:
                    value = obj.attrs.get(name, '')
                except IOError as _exc:
                    value = 'IOError: ' +  str(_exc)
                try:
                    s.append("%s  @%s = %s" % (indentation, name, str(value)))
                except UnicodeDecodeError as _exc:
                    s.append("%s  @%s = %s" % (indentation, name, 'UnicodeDecodeError: ' + str(_exc)))
        return s

    def _renderLinkedObject(self, obj, name, indentation = "  "):
        '''return a [formatted_string] with the name and target of a NeXus linked object'''
        s = []
        s.append("%s%s --> %s" % (indentation, name, obj.attrs['target']))
        return s

    def _renderDataset(self, dset, name, indentation = "  "):
        '''return a [formatted_string] with the contents and structure of a dataset'''
        shape = dset.shape
        # dset.dtype.kind == 'S', nchar = dset.dtype.itemsize
        if self.isNeXus:
            if "target" in dset.attrs:
                if dset.attrs['target'] != dset.name:
                    return ["%s%s --> %s" % (indentation, name, dset.attrs['target'])]
        txType = self._renderDsType(dset)
        txShape = self._renderDsShape(dset)
        s = []
        if dset.dtype.kind == 'S':
            if isinstance(dset.value, numpy.ndarray):
                value = " = %s" % dset.value[0]
            else:
                value = " = %s" % str(dset.value)
            try:
                s += [ "%s%s:%s%s" % (indentation, name, txType, str(value)) ]
            except UnicodeDecodeError as _exc:
                s += [ "%s%s:%s%s" % (indentation, name, txType, 'UnicodeDecodeError: ' + str(_exc)) ]
            s += self._renderAttributes(dset, indentation)
            # dset.dtype.kind == 'S', nchar = dset.dtype.itemsize
        elif dset.dtype.kind == 'O':
            value = " = %s" % str(dset.value)
            s += [ "%s%s:%s%s" % (indentation, name, txType, value) ]
            s += self._renderAttributes(dset, indentation)
        elif shape == (1,):
            value = " = %s" % str(dset[0])
            s += [ "%s%s:%s%s%s" % (indentation, name, txType, 
                                   txShape, value) ]
            s += self._renderAttributes(dset, indentation)
        else:

            if self.array_items_shown > 2:
                value = self._renderArray(dset, indentation + '  ')
                if len(dset.shape) < 2:
                    # show the array inline with the field
                    s += [ "%s%s:%s%s = %s" % (
                            indentation, name, txType, txShape, value) ]
                else:
                    # show multi-D arrays different
                    s += [ "%s%s:%s%s" % (
                            indentation, name, txType, txShape) ]
                    #s += [ "%s%s:%s%s = __array" % (
                    #        indentation, name, txType, txShape) ]
                    # s += [ "%s  %s = %s" % (indentation, "__array", value) ]
            else:
                s += [ "%s%s:%s%s = [ ... ]" % (
                        indentation, name, txType, txShape) ]

            # show these after __array
            s += self._renderAttributes(dset, indentation)
        return s

    def _renderDsType(self, obj):
        ''' get the storage (data) type of the dataset '''
        t = str(obj.dtype)
        # dset.dtype.kind == 'S', nchar = dset.dtype.itemsize
        if obj.dtype.kind == 'S':        # fixed-length string
            t = 'char[%s]' % obj.dtype.itemsize
        elif obj.dtype.kind == 'O':      # variable-length string
            t = 'CHAR'
        if self.isNeXus:
            t = 'NX_' + t.upper()
        return t

    def _renderDsShape(self, obj):
        ''' return the shape of the HDF5 dataset '''
        s = obj.shape
        l = []
        for dim in s:
            l.append(str(dim))
        if l == ['1']:
            result = ""
        else:
            result = "[%s]" % ",".join(l)
        return result

    def _renderArray(self, obj, indentation = '  '):
        ''' nicely format an array up to arbitrary rank '''
        shape = obj.shape
        r = ""
        if len(shape) > 0:
            r = self._renderNdArray(obj, indentation + '  ')
        return r

    def _decideNumShown(self, n):
        ''' determine how many values to show '''
        if self.array_items_shown != None:
            if n > self.array_items_shown:
                n = self.array_items_shown - 2
        return n

    def _renderNdArray(self, obj, indentation = '  '):
        ''' return a list of lower-dimension arrays, nicely formatted '''
        
        def __render(obj, rank, key, indents):
            if rank == 1:
                try:
                    item = obj[key]
                except Exception as _exc:
                    item = str(_exc)
            else:
                # this replaces a lot of code: if rank == ...
                indices = [':' for _ in range(rank)]
                indices[0] = '0'
                indices = ', '.join(indices)
                try:
                    part = eval('obj[%s]' % indices)
                except IOError as _exc:
                    return 'IOError: ' + str(_exc)
                item = self._renderNdArray(part, indents + '  ')    # recursion
            return item

        shape = obj.shape
        rank = len(shape)
        if rank < 1: return None
        n = self._decideNumShown( shape[0] )
        r = []
        for i in range(n):
            r.append( __render(obj, rank, i, indentation + '  ') )
        if n < shape[0]:
            r.append("...")    # skip over most
            r.append( __render(obj, rank, -1, indentation + '  ') ) # last one

        if rank == 1:
            s = str( r )
        else:
            s = "[\n" + indentation + '  '
            s += ("\n" + indentation + '  ').join( r )
            s += "\n" + indentation + "]"
        return s


def isNeXusFile(filename):
    '''
    is `filename` is a NeXus HDF5 file?
    
    Tests if ``filename`` adheres to either
    "Associating plottable data using attributes applied to the **NXdata** group" 
    (needs URL - NeXus manual is not yet ready to provide it)
    or
    "Associating plottable data by name using the ``axes`` attribute"
    (needs URL - again, from NeXus manual)
    or
    "Associating plottable data by dimension number using the ``axis`` attribute"
    (needs URL - again, from NeXus manual)
    '''
    if not os.path.exists(filename):
        return None
    m1 = isNeXusFile_ByNXdataAttrs
    m2 = isNeXusFile_ByAxes
    m3 = isNeXusFile_ByAxisAttr
    # shorter method names to make next line readable
    return m1(filename) or m2(filename) or m3(filename)


def _get_group_niac2014(parent, attribute, nxclass_name):
    '''
    supports the NIAC2014 method:
    
    :param obj parent: instance of h5py.Group or h5py.File
    :param str attribute: this value: ``default``
    :param str nxclass_name: either ``NXentry`` or ``NXdata``
    :return [obj]: list of instances of h5py.Group
    
    Search parent for the group named by the attribute
    or if attribute is not defined, then identify any and all
    groups with the same *nxclass_name*.
    '''
    matches = []
    group = parent.attrs.get(attribute)
    if group is None:
        # Expect that some data files will not write these attributes.
        # Find *any* HDF5 group that has its @NX_class attribute set to ``nxclass_name``.
        for node in parent.values():
            if isNeXusGroup(node, nxclass_name):
                matches.append(node)
    else:
        group = parent.get(group)   # convert str to HDF5 object or a KeyError exception
        if group is not None and isNeXusGroup(group, nxclass_name):
            matches.append(group)
    return matches

def isNeXusFile_ByNXdataAttrs(filename):
    '''
    is `filename` is a NeXus HDF5 file?
    
    This is the "NIAC2014" method.
    In short, verify these NeXus classpaths exist::
    
        /@default={entry_group}
        /{entry_group}:NXentry/@default={data_group}
        /{entry_group}:NXentry/{data_group}:NXdata
        /{entry_group}:NXentry/{data_group}:NXdata/@signal={signal_dataset}
        /{entry_group}:NXentry/{data_group}:NXdata/{signal_dataset}
        /{entry_group}:NXentry/{data_group}:NXdata/@axes=["{axes_dataset1}", ...]
        /{entry_group}:NXentry/{data_group}:NXdata/@{axes_dataset1}_indices=int[]
        ...
    
    where curly braces (``{`` and ``}``) denote that the enclosed name
    is defined in the data file.
    '''
    try:
        f = h5py.File(filename, 'r')
        if not isHdf5File(f):
            f.close()
            return False
        
        # find the NXentry group
        nxentry = _get_group_niac2014(f, 'default', 'NXentry')
        if len(nxentry) == 0:
            return False
        nxentry = nxentry[0]
        
        # find the NXdata group
        nxdata = _get_group_niac2014(nxentry, 'default', 'NXdata')
        if len(nxdata) == 0:
            return False        # no compliant NXdata group identified
        nxdata = nxdata[0]
        
        # find the signal dataset
        signal = nxdata.attrs.get('signal', None)
        if signal is None:
            return False        # no signal attribute
        if isinstance(signal, numpy.ndarray):
            signal = signal[0]
        if signal not in nxdata:
            return False        # no signal dataset
        ds_signal = nxdata[signal]
        if not isNeXusDataset(ds_signal):
            return False        # that HDF5 object is not a NeXus dataset
        
        # Tests for nxdata.attrs['axes'] and nxdata.attrs['{axisname}_indices'] 
        # cannot be robust since we expect some clients simply will not write these.

        f.close()
        return True
    except Exception as _exc:
        pass    # ignore any Exceptions, they mean that result stays "False"
    return False


def isNeXusFile_ByAxes(filename):
    '''
    is `filename` is a NeXus HDF5 file?

    This has been, to date, the most common method in NeXus to define the default plot.
    
    In short, verify this NeXus classpath exists::
    
        /NXentry/NXdata/dataset@signal=1
    
    Tests for the existence of any NXentry group 
    containing any NXdata group containing a single dataset 
    with signal=1 attribute (allows either integer or text representation).
    This is the minimum requirement for a NeXus data file.
    
    This method ignores any exceptions incurred.
    '''
    try:
        f = h5py.File(filename, 'r')
        if not isHdf5File(f):
            f.close()
            return False
        for node0 in f.values():
            if not isNeXusGroup(node0, 'NXentry'):
                continue
            for node1 in node0.values():
                if not isNeXusGroup(node1, 'NXdata'):
                    continue
                signal1_count = 0   # count datasets with signal=1 attribute
                for node2 in node1.values():
                    if isNeXusDataset(node2) and node2.attrs.get('signal', None) in (1, '1'):
                        signal1_count += 1
                if signal1_count == 1:  # ensure only 1 is defined
                    return True
        f.close()
    except:
        pass    # ignore any Exceptions, they mean that result stays "False"
    return False


def isNeXusFile_ByAxisAttr(filename):
    '''
    is `filename` a NeXus HDF5 file?
    
    This is the oldest method in NeXus to define the default plot.
    
    NOTE: **Not implemented yet!** - calls *isNeXusFile_ByAxes()*
    '''
    return isNeXusFile_ByAxes(filename)    # TODO: issue #20: implement this method


def isNeXusGroup(obj, NXtype):
    '''is `obj` a NeXus group?'''
    nxclass = None
    if isHdf5Group(obj):
        nxclass = obj.attrs.get('NX_class', None)
        if isinstance(nxclass, numpy.ndarray):
            nxclass = nxclass[0]
    return nxclass == str(NXtype)


def isNeXusDataset(obj):
    '''is `obj` a NeXus dataset?'''
    return isHdf5Dataset(obj)


def isNeXusLink(obj):
    '''is `obj` linked to another NeXus item?'''
    target = obj.attrs.get('target', '')
    return len(target) > 0 and target != obj.name


def isHdf5File(obj):
    '''is `obj` an HDF5 File?'''
    return isinstance(obj, h5py.File)


def isHdf5Group(obj):
    '''is `obj` an HDF5 Group?'''
    return isinstance(obj, h5py.Group)


def isHdf5Dataset(obj):
    '''is `obj` an HDF5 Dataset?'''
    return isinstance(obj, h5py.Dataset)


def isHdf5Link(obj):
    '''is `obj` an HDF5 Link?'''
    return isinstance(obj, h5py.HardLink)


def isHdf5ExternalLink(obj):
    '''is `obj` an HDF5 ExternalLink?'''
    return isinstance(obj, h5py.ExternalLink)


if __name__ == '__main__':
    print("Start this module using:  python main.py structure ...")
    exit(0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

'''
maintain the local cache of NeXus NXDL and XML Schema files

.. autosummary::
   
   ~NoCacheDirectory
   ~get_nxdl_dir
   ~get_pickle_file_name
   ~write_pickle_file
   ~read_pickle_file
   ~githubMasterInfo
   ~update_NXDL_Cache
   ~qsettings
   ~user_cache_settings
   ~source_cache_settings
   ~SourceCacheSettings
   ~UserCacheSettings
   ~abs_NXDL_filename
   ~get_XML_Schema
   ~get_nxdl_xsd
   ~get_nxdlTypes_xsd


A key component necessary to validate both NeXus data files and 
NXDL class files is a current set of the NXDL definitions.

This code maintains two sets of the definitions.

One is the set provided with the package at installation.  
This set is updated by the developer prior to packaging the 
source for distribution.
Since the source cache is already installed with the package,
it provides a version of the NeXus definitions available for 
fallback use when network access to the GitHub
repository is not available.

The second set is updated into a directory that can be written by 
the user.  This set is updated on demand by the user and only 
when a network connection allows the code to contact the GitHub
source code repository.  The update process will update content 
from the repository.

This code chooses which set of definitions to use, based on the
presence of the *__use_source_cache__ file*.  This file is part
of the source code repository and will be available to 
developers using code from the source code repository.  This file
will not be packaged with the source distribution, thus not present
when users run from a copy of the *punx* package installed from PyPI
(or other).

.. rubric:: Public Interface

:settings object:     :meth:`~punx.cache.qsettings`
:get new NXDL definitions from GitHub:     :meth:`~punx.cache.update_NXDL_Cache`
'''

import cPickle as pickle
import lxml
import os
import requests.packages.urllib3
from requests.packages.urllib3.exceptions import InsecureRequestWarning
import StringIO
import zipfile
from PyQt4 import QtCore

import nxdlstructure
import settings
import __init__


orgName = __init__.__settings_organization__
appName = __init__.__settings_package__

PKG_DIR = os.path.abspath(os.path.dirname(__file__))
SOURCE_CACHE_ROOT = os.path.join(PKG_DIR, __init__.CACHE_SUBDIR)

SOURCE_CACHE_KEY_FILE = '__use_source_cache__'
USE_SOURCE_CACHE = os.path.exists(os.path.join(PKG_DIR, SOURCE_CACHE_KEY_FILE))

NXDL_SCHEMA_FILE = 'nxdl.xsd'
NXDL_TYPES_SCHEMA_FILE = 'nxdlTypes.xsd'

NXDL_NAMESPACE = 'http://definition.nexusformat.org/nxdl/3.1'
XSD_NAMESPACE = 'http://www.w3.org/2001/XMLSchema'
NX_DICT = dict(xs=XSD_NAMESPACE, nx=NXDL_NAMESPACE)


__singleton_cache_settings_source__ = None
__singleton_cache_settings_user__ = None
__singleton_settings__ = None
__singleton_xml_schema__ = None
__singleton_nxdl_xsd__ = None
__singleton_nxdlTypes_xsd__ = None


class NoCacheDirectory(Exception): 
    'custom exception: no cache directory was found'
    pass


def get_nxdl_dir():
    '''
    the path of the directory with the files containing NeXus definitions
    
    Note:  This directory, and the files it contains, are **only** used during
    the process of updating the cache and then writing the pickle file.
    '''
    if USE_SOURCE_CACHE:
        cache_dir = SOURCE_CACHE_ROOT
    else:
        cache_dir = qsettings().cache_dir()
    path = os.path.abspath(os.path.join(cache_dir, __init__.NXDL_CACHE_SUBDIR))
    return path


def get_pickle_file_name(path, use_fallback=True):
    '''
    The pickle file holds all known NXDL classes, parsed into Python data structures
    '''
    pfile = os.path.join(path, __init__.PICKLE_FILE)
    if use_fallback and not USE_SOURCE_CACHE and not os.path.exists(pfile):
        # user cache has no pickle file, fall back to source cache
        __init__.LOG_MESSAGE('using source cache pickle file', __init__.DEBUG)
        pfile = os.path.join(SOURCE_CACHE_ROOT, __init__.PICKLE_FILE)
    return os.path.abspath(pfile)


def write_pickle_file(info, path):
    '''
    write the parsed nxdl_dict and info to a Python pickle file
    '''
    info['pickle_file'] = get_pickle_file_name(path, use_fallback=False)
    for k, v in info.items():
        msg = 'info[%s] = %s' % (k, str(v))
        __init__.LOG_MESSAGE(msg, __init__.DEBUG)
    msg = 'update pickle file: ' + os.path.abspath(info['pickle_file'])
    __init__.LOG_MESSAGE(msg, __init__.INFO)

    nxdl_dict = nxdlstructure.get_NXDL_specifications()
    pickle_data = dict(nxdl_dict=nxdl_dict, info=info)
    pickle.dump(pickle_data, open(info['pickle_file'], 'wb'))


def read_pickle_file(pfile, sha):
    '''
    read the parsed nxdl_dict and info from a Python pickle file
    '''
    __init__.LOG_MESSAGE('read pickle file', __init__.DEBUG)
    pickle_data = pickle.load(open(pfile, 'rb'))
    if 'info' in pickle_data:
        # any other tests to qualify this?
        if sha == pickle_data['info']['git_sha']:   # declare victory!
            # do not need to return ``info`` since it matches
            __init__.LOG_MESSAGE('matching SHA', __init__.DEBUG)
            return pickle_data['nxdl_dict']
    __init__.LOG_MESSAGE('SHA does not match', __init__.DEBUG)
    return None


def githubMasterInfo(org, repo):
    '''
    get information about the repository master branch
    
    :returns: dict (as below) or None if could not get info
    
    ========  ================================================
    key       meaning
    ========  ================================================
    git_time  ISO-8601-compatible timestamp from GitHub
    git_sha   hash tag of latest GitHub commit
    zip_url   URL of downloadable ZIP file
    ========  ================================================
    '''
    # get repository information via GitHub API
    url = 'https://api.github.com/repos/%s/%s/commits' % (org, repo)
    
    msg = 'disabling warnings about GitHub self-signed https certificates'
    __init__.LOG_MESSAGE(msg, __init__.DEBUG)
    requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
    __init__.LOG_MESSAGE('get repo info: ' + str(url), __init__.INFO)
    for _retry in range(__init__.GITHUB_RETRY_COUNT):
        try:
            r = requests.get(url, verify=False)
        except requests.exceptions.ConnectionError as _exc:
            # see: http://docs.python-requests.org/en/master/user/quickstart/#errors-and-exceptions
            # see: http://docs.python-requests.org/en/master/api/#id1
            __init__.LOG_MESSAGE('ConnectionError from ' + str(url), __init__.ERROR)
            return None
        else:
            break
        __init__.LOG_MESSAGE('retry to get repo info: ' + str(url), __init__.WARN)

    latest = r.json()[0]
    sha = latest['sha']
    iso8601 = latest['commit']['committer']['date']
    zip_url = 'https://github.com/%s/%s/archive/master.zip' % (org, repo)
    
    __init__.LOG_MESSAGE('git sha: ' + str(sha), __init__.INFO)
    __init__.LOG_MESSAGE('git iso8601: ' + str(iso8601), __init__.INFO)
    
    return dict(git_sha=sha, git_time=iso8601, zip_url=zip_url)


def __get_github_info__():
    '''
    check with GitHub for any updates
    '''
    return githubMasterInfo(__init__.GITHUB_NXDL_ORGANIZATION, 
                            __init__.GITHUB_NXDL_REPOSITORY)


def update_NXDL_Cache(force_update=False):
    '''
    update the cache of NeXus NXDL files
    
    :param bool force_update: (optional, default: False) update if GitHub is available
    '''
    msg = 'update_NXDL_Cache(): force_update=' + str(force_update)
    __init__.LOG_MESSAGE(msg, __init__.DEBUG)
    info = __get_github_info__()    # check with GitHub first
    if info is None:
        __init__.LOG_MESSAGE('GitHub not available', __init__.INFO)
        return

    # proceed only if GitHub info was available
    qset = qsettings()
    info['file'] = str(qset.fileName())

    different_sha = str(info['git_sha']) != str(qset.getKey('git_sha'))
    different_git_time = str(info['git_time']) != str(qset.getKey('git_time'))
    nxdl_subdir_exists = os.path.exists(get_nxdl_dir())

    could_update = different_sha or different_git_time or not nxdl_subdir_exists
    updating = could_update or force_update
    
    if not updating:
        __init__.LOG_MESSAGE('not updating NeXus definitions files', __init__.INFO)
        return

    path = qset.cache_dir()
    msg = 'updating NeXus definitions files in directory: ' + os.path.abspath(path)
    __init__.LOG_MESSAGE(msg, __init__.INFO)

    # download the repository ZIP file 
    url = info['zip_url']
    msg = 'disabling warnings about GitHub self-signed https certificates'
    __init__.LOG_MESSAGE(msg, __init__.DEBUG)
    requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
    __init__.LOG_MESSAGE('download: ' + str(url), __init__.INFO)
    for _retry in range(__init__.GITHUB_RETRY_COUNT):
        try:
            content = requests.get(url, verify=False).content
        except requests.exceptions.ConnectionError as _exc:
            msg = 'ConnectionError from ' + str(url)
            msg += ', ' + str(_exc)
            __init__.LOG_MESSAGE(msg, __init__.ERROR)
            return None
        else:
            break
        __init__.LOG_MESSAGE('retry download: ' + str(url), __init__.WARN)

    buf = StringIO.StringIO(content)
    zip_content = zipfile.ZipFile(buf)
    # TODO: How to save this ZIP to disk? not needed when using pickle file
    #local_zip_file_name = os.path.join(path, url.split('/')[-1])
    #open(local_zip_file_name, 'w').write(content)
    
    # extract the NXDL (XML, XSL, & XSD) files to the path
    msg = 'extract ZIP to directory: ' + os.path.abspath(path)
    __init__.LOG_MESSAGE(msg, __init__.INFO)
    NXDL_categories = 'base_classes applications contributed_definitions'
    for item in zip_content.namelist():
        parts = item.rstrip('/').split('/')
        if len(parts) == 2:             # get the XML Schema files
            if os.path.splitext(parts[1])[-1] in ('.xsd',):
                zip_content.extract(item, path)
                msg = 'extracted: ' + os.path.abspath(item)
                __init__.LOG_MESSAGE(msg, __init__.DEBUG)
        elif len(parts) == 3:         # get the NXDL files
            if parts[1] in NXDL_categories.split():
                if os.path.splitext(parts[2])[-1] in ('.xml .xsl'.split()):
                    zip_content.extract(item, path)
                    msg = 'extracted: ' + os.path.abspath(item)
                    __init__.LOG_MESSAGE(msg, __init__.DEBUG)
    
    # optimization: write the parsed NXDL specifications to a file
    if force_update:
        # force the pickle file to be re-written
        pfile = get_pickle_file_name(path, use_fallback=False)
        msg = 'force pickle file update'
        __init__.LOG_MESSAGE(msg, __init__.DEBUG)
        if os.path.exists(pfile):
            os.remove(pfile)
    write_pickle_file(info, path)

    if force_update:
        # force the .ini file to be re-written
        __init__.LOG_MESSAGE('force .ini file update', __init__.DEBUG)
        key = 'pickle_file'
        v =  qset.getKey(key)
        qset.setKey(key, 'update forced')
        qset.setKey(key, v)
        for k in ('git_sha', 'git_time'):
            qset.setKey('___global___/'+k, info[k])
    __init__.LOG_MESSAGE('update .ini file: ' + qset.fileName(), __init__.INFO)
    qset.updateGroupKeys(info)


def qsettings():
    '''
    return the QSettings instance, chosen from user or source cache
    '''
    global __singleton_settings__
    if __singleton_settings__ is None:
        if USE_SOURCE_CACHE:
            qset = source_cache_settings()
        else:
            qset = user_cache_settings()
        __singleton_settings__ = qset
    if not os.path.exists(__singleton_settings__.cache_dir()):
        raise NoCacheDirectory('no cache found')
    return __singleton_settings__


def user_cache_settings():
    '''manage the user cache info file as an .ini file'''
    global __singleton_cache_settings_user__
    if __singleton_cache_settings_user__ is None:
        try:
            qset = UserCacheSettings()
        except:
            # fall back to source cache if cannot access user cache 
            qset = SourceCacheSettings()
        __singleton_cache_settings_user__ = qset
    return __singleton_cache_settings_user__


def source_cache_settings():
    '''manage the source cache info file as an .ini file'''
    global __singleton_cache_settings_source__
    if __singleton_cache_settings_source__ is None:
        qset = SourceCacheSettings()
        __singleton_cache_settings_source__ = qset
    return __singleton_cache_settings_source__


class SourceCacheSettings(QtCore.QSettings, settings.QSettingsMixin):
    '''
    manage the source cache settings file as an .ini file using QSettings
    '''
    
    def __init__(self):
        if USE_SOURCE_CACHE and not os.path.exists(SOURCE_CACHE_ROOT):
            os.mkdir(SOURCE_CACHE_ROOT)
        path = os.path.join(SOURCE_CACHE_ROOT, 
                            __init__.SOURCE_CACHE_SETTINGS_FILENAME)
        QtCore.QSettings.__init__(self, path, QtCore.QSettings.IniFormat)
        self.init_global_keys()


class UserCacheSettings(QtCore.QSettings, settings.QSettingsMixin):
    '''
    manage and preserve default settings for this application using QSettings
    
    Use the .ini file format and save under user directory

    :see: http://doc.qt.io/qt-4.8/qsettings.html
    '''
    
    def __init__(self):
        QtCore.QSettings.__init__(self, 
                                  QtCore.QSettings.IniFormat, 
                                  QtCore.QSettings.UserScope, 
                                  orgName, 
                                  appName)
        path = self.cache_dir()
        if not USE_SOURCE_CACHE and not os.path.exists(path):
            os.mkdir(path)
        self.init_global_keys()


def abs_NXDL_filename(file_name):
    '''return absolute path to file_name, within NXDL directory'''
    absolute_name = os.path.abspath(os.path.join(get_nxdl_dir(), file_name))
    msg = 'file does not exist: ' + absolute_name
    if os.path.exists(absolute_name):
        __init__.LOG_MESSAGE('user cache: ' + absolute_name, __init__.DEBUG)
        return absolute_name
    if not USE_SOURCE_CACHE:
        __init__.LOG_MESSAGE('try source cache for: ' + file_name, __init__.DEBUG)
        path = os.path.join(SOURCE_CACHE_ROOT, __init__.NXDL_CACHE_SUBDIR)
        absolute_name = os.path.abspath(os.path.join(path, file_name))
        if os.path.exists(absolute_name):
            __init__.LOG_MESSAGE('source cache: ' + absolute_name, __init__.DEBUG)
            return absolute_name
        msg += '\t AND not found in source cache either!  Report this problem to the developer.'
    raise __init__.FileNotFound(msg)


def get_XML_Schema():
    '''
    parse & cache the XML Schema file (nxdl.xsd) as an **XML Schema** only once
    
    Uses :class:`lxml.etree.XMLSchema` and :meth:`~get_nxdl_xsd`
    '''
    global __singleton_xml_schema__

    if __singleton_xml_schema__ is None:
        __singleton_xml_schema__ = lxml.etree.XMLSchema(get_nxdl_xsd())

    return __singleton_xml_schema__


def get_nxdl_xsd():
    '''
    parse and cache the XML Schema file (nxdl.xsd) as an XML document only once
    
    Uses :meth:`lxml.etree.parse`
    '''
    global __singleton_nxdl_xsd__

    if __singleton_nxdl_xsd__ is None:
        try:
            xsd_file_name = abs_NXDL_filename(NXDL_SCHEMA_FILE)
        except __init__.FileNotFound as _exc:
            raise __init__.SchemaNotFound(_exc)

        if not os.path.exists(xsd_file_name):
            msg = 'Could not find XML Schema file: ' + xsd_file_name
            raise __init__.SchemaNotFound(msg)
    
        __singleton_nxdl_xsd__ = lxml.etree.parse(xsd_file_name)

    return __singleton_nxdl_xsd__


def get_nxdlTypes_xsd():
    '''
    parse and cache the XML Schema file (nxdlTypes.xsd) as an XML document only once
    '''
    global __singleton_nxdlTypes_xsd__

    if __singleton_nxdlTypes_xsd__ is None:
        xsd_file_name = abs_NXDL_filename(NXDL_TYPES_SCHEMA_FILE)

        if not os.path.exists(xsd_file_name):
            msg = 'Could not find XML Schema file: ' + xsd_file_name
            raise IOError(msg)
    
        __singleton_nxdlTypes_xsd__ = lxml.etree.parse(xsd_file_name)

    return __singleton_nxdlTypes_xsd__


if __name__ == '__main__':
#     update_NXDL_Cache(True)
# #     # print user_cache_settings().fileName()
# #     # print source_cache_settings().fileName()
    print("Start this module using:  python main.py update ...")
    exit(0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE, distributed with this software.
#-----------------------------------------------------------------------------

from setuptools import setup, find_packages
import os
import re
import sys

# pull in some definitions from the package's __init__.py file
sys.path.insert(0, os.path.join('src', ))
import punx


verbose=1
long_description = open('README.rst', 'r').read()


setup (name =  punx.__package_name__,        # punx
       license          = punx.__license__,
       version          = punx.__version__,
       description      = punx.__description__,
       long_description = long_description,
       author           = punx.__author_name__,
       author_email     = punx.__author_email__,
       url              = punx.__url__,
       #download_url     = punx.__download_url__,
       keywords         = punx.__keywords__,
       platforms        = 'any',
       install_requires = punx.__install_requires__,
       package_dir      = {'': 'src'},
       packages         = ['punx', ],
       #packages=find_packages(),
       package_data     = {
                            'punx': ['cache/*.p', 
                                     'cache/*.ini', 
                                     'cache/*.zip', 
                                     'cache/*/*.xsd',
                                     'cache/*/*/*.xml',
                                     'cache/*/*/*.xsl',
                                     'data/writer_*.hdf5',
                                     'LICENSE.txt',
                                     'VERSION',
                                     ],
                         },
       classifiers      = punx.__classifiers__,
       entry_points     = {
            # create & install scripts in <python>/bin
            'console_scripts': [
                'punx=punx.main:main',
			],
            #'gui_scripts': [],
       },
      )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

'''
document program history events in a log file

.. autosummary::
   
   ~Logger
   ~to_console

'''

import datetime
import logging
import os
import socket
import sys

import __init__


class Logger(object):
    '''
    use Python logging package to record program history

    :param str log_file: name of file to store history
    :param enum level: logging interest level (default=__init__.INFO, no logs = -1)
    '''

    def __init__(self, log_file=None, level=None):
        if level is None:
            level = __init__.INFO
        self.level = level
        if level == __init__.CONSOLE_ONLY:
            # this means: only write ALL log messages to the console
            self.log_file = None
        else:
            if log_file is None:
                ymd = str(_now()).split()[0]
                pid = os.getpid()
                # current working directory?
                log_file = '-'.join((__init__.__package_name__, ymd, str(pid) + '.log'))
            self.log_file = os.path.abspath(log_file)
            logging.basicConfig(filename=log_file, level=level)

        self.history = ''
        self.filename = os.path.basename(sys.argv[0])
        self.pid = os.getpid()

        __init__.LOG_MESSAGE = self.add
        self.first_logs()

    def add(self, message, interest=None):
        '''
        log a message or report from the application

        :param str message: words to be logged
        :param int interest: interest level of this message (default: logging.INFO)
        '''
        if interest is None:
            interest = __init__.INFO
        if interest < self.level:
            return

        timestamp = _now()
        text = "(%d,%s,%s) %s" % (self.pid, self.filename, timestamp, message)

        if self.level == __init__.CONSOLE_ONLY:
            print(text)
        else:
            logging.log(interest, text)

        if len(self.history) != 0:
            self.history += '\n'
        self.history += text

        return text

    def first_logs(self):
        '''
        first logging information after log file has been defined
        '''
        user = os.environ.get('LOGNAME', None) or os.environ.get('USERNAME', None)
        if self.level == __init__.CONSOLE_ONLY:
            interest = 'no logging'
        else:
            interest = logging.getLevelName(self.level)

        self.add("startup")
        self.add("log_file         = " + str(self.log_file))
        self.add("interest level   = " + interest)
        self.add("user             = " + user)
        self.add("host             = " + socket.gethostname())
        self.add("program          = " + sys.argv[0])
        self.add("program filename = " + self.filename)
        self.add("PID              = " + str(self.pid))


def to_console(message, interest=None):
    '''
    used when *only* logging output to the console (not using the logging package)
    '''
    if interest >= __init__.DEFAULT_LOG_LEVEL:
        status = logging.getLevelName(interest) + ':'
        print(status, message)


def ignore_logging():
    '''
    used during unit testing
    '''
    __init__.DEFAULT_LOG_LEVEL = 999999
    __init__.LOG_MESSAGE = to_console


def _now():
    ''' '''
    return datetime.datetime.now()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

'''
Python Utilities for NeXus HDF5 files

main user interface file

.. rubric:: Usage

::

    usage: punx [-h] [-v] {hierarchy,show,structure,update,validate} ...
    
    Python Utilities for NeXus HDF5 files URL: http://punx.readthedocs.io
    v0.0.1+4.gff00892.dirty
    
    optional arguments:
      -h, --help            show this help message and exit
      -v, --version         show program's version number and exit
    
    subcommands:
      valid subcommands
    
      {demonstrate,structure,update,validate}
        demonstrate         demonstrate HDF5 file validation
        structure           show structure of HDF5 or NXDL file
        update              update the local cache of NeXus definitions
        validate            validate a NeXus file

.. autosummary::
   
   ~main
   ~MyArgumentParser
   ~parse_command_line_arguments
   ~interceptor_logfile
   ~func_demo
   ~func_hierarchy
   ~func_show
   ~func_structure
   ~func_update
   ~func_validate

'''

import argparse
import os
import sys

import __init__
import finding
import logs


CONSOLE_LOGGING_DEFAULT_CHOICE = '__console__'


# :see: https://docs.python.org/2/library/argparse.html#sub-commands
# obvious 1st implementations are h5structure and update


def exit_message(msg, status=None, exit_code=1):
    '''
    exit this code with a message and a status
    
    :param str msg: text to be reported
    :param int status: 0 - 50 (default: __init__.ERROR = 40)
    :param int exit_code: 0: no error, 1: error (default)
    '''
    if status is None:
        status = __init__.ERROR
    __init__.LOG_MESSAGE(msg, status)
    if __init__.LOG_MESSAGE != logs.to_console:
        print(msg)
    exit(exit_code)


def func_demo(args):
    '''
    show what **punx** can do
    
    .. index:: demo

    Internally, runs these commands::

        punx validate <source_directory>/data/writer_1_3.hdf5
        punx structure <source_directory>/data/writer_1_3.hdf5

    .. index:: cache update

    If you get an error message that looks like this one (line breaks added here for clarity::

        punx.cache.FileNotFound: file does not exist:
        /Users/<username>/.config/punx/definitions-master/nxdl.xsd
        AND not found in source cache either!  Report this problem to the developer.

    then you will need to update your local cache of the NeXus definitions.
    Use this command to update the local cache::

        punx update

    '''
    path = os.path.dirname(__file__)
    args.infile = os.path.abspath(os.path.join(path, 'data', 'writer_1_3.hdf5'))

    print('console> punx validate ' + args.infile)
    args.report = ','.join(sorted(finding.VALID_STATUS_DICT.keys()))
    func_validate(args)
    del args.report

    print('console> punx structure ' + args.infile)
    import h5structure
    mc = h5structure.h5structure(args.infile)
    #    :param bool show_attributes: display attributes in output
    show_attributes=True
    mc.array_items_shown = 5
    print('\n'.join(mc.report(show_attributes)))


def func_hierarchy(args):
    url = 'http://punx.readthedocs.io/en/latest/analyze.html'
    print('A chart of the NeXus hierarchy is in the **punx** documentation.')
    print('see: ' + url)
    # TODO: issue #1 & #10 show NeXus base class hierarchy from a given base class


def func_show(args):
    print('still in development -- not implemented yet')
    print(args)


def func_structure(args):
    if args.infile.endswith('.nxdl.xml'):
        import nxdlstructure
        nxdl = nxdlstructure.NX_definition(args.infile)
        print(nxdl.render())
    else:
        import h5structure
        
        #    :param int limit: maximum number of array items to be shown (default = 5)
        limit = args.max_array_items
        #    :param bool show_attributes: display attributes in output
        show_attributes=True

        try:
            mc = h5structure.h5structure(os.path.abspath(args.infile))
        except __init__.FileNotFound:
            exit_message('File not found: ' + args.infile)
        mc.array_items_shown = limit
        try:
            report = mc.report(show_attributes)
        except __init__.HDF5_Open_Error:
            exit_message('Could not open as HDF5: ' + args.infile)
        print('\n'.join(report or ''))


def func_update(args):
    import cache
    cache.update_NXDL_Cache(force_update=args.force)


def func_validate(args):
    import validate

    if args.infile.endswith('.nxdl.xml'):
        result = validate.validate_xml(args.infile)
        if result is None:
            print(args.infile, ' validates')
    else:
        try:
            validator = validate.Data_File_Validator(args.infile)
        except __init__.FileNotFound:
            exit_message('File not found: ' + args.infile)
        except __init__.HDF5_Open_Error:
            exit_message('Could not open as HDF5: ' + args.infile)
        except __init__.SchemaNotFound as _exc:
            exit_message(str(_exc))
    
        # determine which findings are to be reported
        report_choices, trouble = [], []
        for c in args.report.upper().split(','):
            if c in finding.VALID_STATUS_DICT:
                report_choices.append(finding.VALID_STATUS_DICT[c])
            else:
                trouble.append(c)
        if len(trouble) > 0:
            msg = 'invalid choice(s) for *--report* option: '
            msg += ','.join(trouble)
            msg += '\n'
            msg += '\t' + 'available choices: '
            msg += ','.join(sorted(finding.VALID_STATUS_DICT.keys()))
            exit_message(msg)

        # run the validation
        validator.validate()

        # report the findings from the validation
        import cache
        qset = cache.qsettings()
        print('Validate file ' + os.path.basename(validator.fname))
        print(':NXDL cache: ' + cache.get_nxdl_dir())
        print(':NXDL GIT sha: ' + qset.getKey('git_sha'))
        print(':NXDL GIT date/time: ' + qset.getKey('git_time'))
        print(':validation results shown: ' + ' '.join(sorted(map(str, report_choices))))
        print('')
        print('Validation findings')
        print(validator.report_findings(report_choices))
        
        print('summary statistics')
        print(validator.report_findings_summary())


class MyArgumentParser(argparse.ArgumentParser):
    '''
    override standard ArgumentParser to enable shortcut feature
    
    stretch goal: permit the first two char (or more) of each subcommand to be accepted
    # ?? http://stackoverflow.com/questions/4114996/python-argparse-nargs-or-depending-on-prior-argument?rq=1
    '''
    
    def parse_args(self, args=None, namespace=None):
        '''
        permit the first two char (or more) of each subcommand to be accepted
        '''
        if args is None and len(sys.argv) > 1 and not sys.argv[1].startswith('-'):
            # TODO: issue #8: make more robust for variations in optional commands
            sub_cmd = sys.argv[1]
            # make a list of the available subcommand names
            choices = []
            for g in self._subparsers._group_actions:
                if isinstance(g, argparse._SubParsersAction):
                    #choices = g._name_parser_map.keys()
                    choices = g.choices.keys()
                    break
            if len(choices) > 0 and sub_cmd not in choices:
                if len(sub_cmd) < 2:
                    msg = 'subcommand too short, must match first 2 or more characters, given: %s'
                    self.error(msg % ' '.join(sys.argv[1:]))
                # look for any matches
                matches = [c for c in choices if c.startswith(sub_cmd)]
                # validate the match is unique
                if len(matches) == 0:
                    msg = 'subcommand unrecognized, given: %s'
                    self.error(msg % ' '.join(sys.argv[1:]))
                elif len(matches) > 1:
                    msg = 'subcommand ambiguous (matches: %s)' % ' | '.join(matches)
                    msg += ', given: %s'
                    self.error(msg % ' '.join(sys.argv[1:]))
                else:
                    sub_cmd = matches[0]
                # re-assign the subcommand
                sys.argv[1] = sub_cmd
        return argparse.ArgumentParser.parse_args(self, args, namespace)


def parse_command_line_arguments():
    '''process command line'''
    doc = __doc__.strip().splitlines()[0]
    doc += '\n  version: ' + __init__.__version__
    doc += '\n  release: ' + __init__.__release__
    doc += '\n  URL: ' + __init__.__url__
    epilog = 'Note: It is only necessary to use the first two (or'
    epilog += ' more) characters of any subcommand, enough that the'
    epilog += ' abbreviation is unique. '
    epilog += ' Such as: ``demonstrate`` can be abbreviated to'
    epilog += ' ``demo`` or even ``de``.'
    p = MyArgumentParser(prog=__init__.__package_name__, 
                                     description=doc,
                                     epilog=epilog)

    p.add_argument('-v', 
                    '--version', 
                    action='version', 
                    version=__init__.__version__)
    
    def add_logging_argument(subp):
        '''
        common code to add option for logging program output
        '''
        import logging
        help_text = 'log output to file (default: no log file)'
        subp.add_argument('-l', '--logfile',
                       default=CONSOLE_LOGGING_DEFAULT_CHOICE,
                       nargs='?',
                       help=help_text)
        
        level = __init__.DEFAULT_LOG_LEVEL
        help_text = 'logging interest level (%d - %d), default=%d (%s)'
        help_text = help_text % (__init__.NOISY,
                                 __init__.CRITICAL,
                                 __init__.DEFAULT_LOG_LEVEL,
                                 logging.getLevelName(level)
                                 )
        subp.add_argument('-i', '--interest',
                       default=level,
                       type=int,
                       #choices=range(1,51),
                       help=help_text)

    # TODO: issue #9, stretch goal: GUI for any of this
    # p.add_argument('-g', 
    #                     '--gui', 
    #                     help='graphical user interface (TBA)')

    subcommand = p.add_subparsers(title='subcommand', description='valid subcommands',)
    
    
    ### subcommand: demo
    p_demo = subcommand.add_parser('demonstrate', help='demonstrate HDF5 file validation')
    add_logging_argument(p_demo)
    p_demo.set_defaults(func=func_demo)


#     ### subcommand: hierarchy
#     # TODO: issue #1 & #10
#     help_text = 'show NeXus base class hierarchy from a given base class'
#     p_hierarchy = subcommand.add_parser('hierarchy',  help=help_text)
#     p_hierarchy.set_defaults(func=func_hierarchy)
#     #p_hierarchy.add_argument('something', type=bool, help='something help_text')


    ### subcommand: show
#     # TODO: issue #11
#     p_show = subcommand.add_parser('show', help='show program information (about the cache)')
#     p_show.set_defaults(func=func_show)
#     # p_show.add_argument('details', type=bool, help='details help_text')


    ### subcommand: structure
    help_text = 'show structure of HDF5 or NXDL file'
    p_structure = subcommand.add_parser('structure', help=help_text)
    p_structure.set_defaults(func=func_structure)
    p_structure.add_argument('infile', help="HDF5 or NXDL file name")
    p_structure.add_argument('-a', 
                        action='store_false', 
                        default=True,
                        dest='show_attributes',
                        help='Do not print attributes of HDF5 file structure')
    help_text = 'maximum number of array items to be shown'
    p_structure.add_argument('-m', '--max_array_items',
                       default=5,
                       type=int,
                       #choices=range(1,51),
                       help=help_text)
    add_logging_argument(p_structure)


    ### subcommand: update
    help_text = 'update the local cache of NeXus definitions'
    p_update = subcommand.add_parser('update', help=help_text)
    p_update.set_defaults(func=func_update)
    p_update.add_argument('-f', '--force', 
                               action='store_true', 
                               default=False, 
                               help='force update (if GitHub available)')
    add_logging_argument(p_update)


    ### subcommand: validate
    p_validate = subcommand.add_parser('validate', help='validate a NeXus file')
    p_validate.add_argument('infile', help="HDF5 or NXDL file name")
    p_validate.set_defaults(func=func_validate)
    reporting_choices = ','.join(sorted(finding.VALID_STATUS_DICT.keys()))
    help_text = 'select which validation findings to report, choices: '
    help_text += reporting_choices
    p_validate.add_argument('--report', default=reporting_choices, help=help_text)
    add_logging_argument(p_validate)

    return p.parse_args()


def interceptor_logfile(args):
    '''
    special handling for subcommands with a *logfile* option
    '''
    if 'logfile' in args:
        if args.logfile == CONSOLE_LOGGING_DEFAULT_CHOICE:
            __init__.DEFAULT_LOG_LEVEL = args.interest
            __init__.LOG_MESSAGE = logs.to_console
            __init__.LOG_MESSAGE('logging output to console only', __init__.DEBUG)
        else:
            lo = __init__.NOISY
            hi = __init__.CRITICAL
            args.interest = max(lo, min(hi, args.interest))
            _log = logs.Logger(log_file=args.logfile, level=args.interest)
        __init__.LOG_MESSAGE('sys.argv: ' + ' '.join(sys.argv), __init__.DEBUG)
        __init__.LOG_MESSAGE('args: ' + str(args), __init__.DEBUG)


def main():
    args = parse_command_line_arguments()
    
    # special handling for logging program output
    interceptor_logfile(args)

    args.func(args)


if __name__ == '__main__':
    main()
